import numpy as np
import pandas as pd
import torch as t
from sklearn.preprocessing import StandardScaler
from torch.utils.data import Dataset

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.paths import *


class NarrowCVAEDataset(Dataset):

    def __init__(self, cfg: MasterConfig, train: bool):
        """
        Initialize.
        """

        self.cfg = cfg
        self.train = train
        self.data_ratio = cfg.sl.general.data_ratio
        self.path = None
        self.x = None
        self.y = None

        self._setup()

    def __getitem__(self, index: int) -> dict:
        """
        Get sample.
        :param index: sample index.
        :return: sample (x and y fields).
        """

        sample = dict()
        sample['x'] = t.tensor(self.x[index, :], dtype=t.float32, device='cuda')
        sample['y'] = t.tensor(self.y[index, :], dtype=t.float32, device='cuda')

        return sample

    def __len__(self):
        """
        Len of dataset.
        :return: len of dataset.
        """

        return self.x.shape[0]

    def _setup(self):
        """
        Set up dataset.
        """

        self._setup_path()
        self._setup_arrays()

    def _setup_path(self):
        """
        Setup appropriate csv path.
        """

        if self.train:
            self.path = Paths().data_sl_narrow_train_csv
        else:
            self.path = Paths().data_sl_narrow_val_csv

    def _setup_arrays(self):
        """
        Set up data into two arrays x (states) and y (conditions).
        """

        # Read from master file
        arr = pd.read_csv(self.path, sep=',').to_numpy()

        # Constants
        y_dim = 2 * self.cfg.sl.dim.state + self.cfg.sl.dim.obstacle
        x_dim = self.cfg.sl.dim.state

        # Split into x and y taking the ratio (assume data start at index 1 and is y then x)
        stop_at = int(self.cfg.sl.general.data_ratio * len(arr))
        self.x = arr[:stop_at, -x_dim:]
        self.y = arr[:stop_at, 1:(y_dim + 1)]


class DistanceOrdinalDataset(Dataset):

    def __init__(self,
                 cfg: MasterConfig,
                 split_type: str,
                 scaler: None | StandardScaler = None,
                 timesteps_range: None | list[int, int] | str = None):
        """
        Initialize.
        """

        self.cfg = cfg
        self.split_type = split_type
        self.scaler = scaler
        self.timesteps_range = timesteps_range

        self.x = None
        self.y = None

        self._setup()

    def __getitem__(self, index):
        """
        Get sample.
        :param index: sample index.
        :return: sample as namedtuple (x and y fields).
        """

        x = t.tensor(self.x[index, :], dtype=t.float32, device="cuda")
        y = t.tensor(self.y[index], dtype=t.float32, device="cuda")

        return [x, y]

    def __len__(self):
        """
        Len of dataset.
        :return:
        """

        return self.x.shape[0]

    def _setup(self):
        """
        Set up.
        """

        self._setup_path()
        self._setup_arrays()
        self._setup_scaling()

    def _setup_arrays(self):
        """
        Set up data into two arrays x (states) and y (conditions).
        """

        col_names = ['x', 'y', 'sin_theta', 'cos_theta', 'success', 'time', 'timesteps', 'length']

        # Read and remove some columns
        df = pd.read_csv(self.path, names=col_names, index_col=None)
        df.drop(columns=['success', 'time', 'length'], axis=1, inplace=True)

        # Remove above cutoff
        df = df[df['timesteps'] < self.cfg.rrt.ordinal.max_timesteps]

        # Get or set min and max timestep values
        if self.timesteps_range is None:
            self.timesteps_range = [df['timesteps'].min(), df['timesteps'].max()]

        # Scale target to range
        df['timesteps'] = df['timesteps'] - self.timesteps_range[0]

        # Convert to numpy array and get only part of the data        
        stop_at = int(self.cfg.rrt.ordinal.data_ratio * len(df))
        self.x = df.iloc[:stop_at, :-1].to_numpy()
        self.y = df.iloc[:stop_at, -1].to_numpy().astype(int)

    def _setup_path(self):
        """
        Set up appropriate csv path.
        """

        match self.split_type:
            case 'train':
                self.path = Paths().data_rl_distance / self.cfg.rrt.names.rl / "train.csv"
            case 'val':
                self.path = Paths().data_rl_distance / self.cfg.rrt.names.rl / "val.csv"
            case 'test':
                self.path = Paths().data_rl_distance / self.cfg.rrt.names.rl / "test.csv"
            case _:
                raise NotImplementedError

    def _setup_scaling(self):
        """
        Set up data scaling.
        """

        # Instantiate and fit scaler for training
        if self.scaler is None:
            self.scaler = StandardScaler().fit(self.x)

        # Transform x data
        self.x = self.scaler.transform(self.x)




from torch.utils.data import DataLoader


class FixedDataLoader(object):
    def __init__(self, dataloader):
        # Check input
        if not isinstance(dataloader, DataLoader):
            msg = (
                "The input used to instantiate FixedDataLoader should be a"
                " DataLoader from `torch.utils.data`."
            )
            raise ValueError(msg)

        self.elem_list = []
        for _, elem in enumerate(dataloader):
            self.elem_list.append(elem)

    def __getitem__(self, index):
        return self.elem_list[index]

    def __len__(self):
        return len(self.elem_list)



import pytest

from torchensemble.utils.logging import set_logger


def _record(logger):
    logger.debug("Debug!")
    logger.info("Info!")
    logger.warning("Warning!")
    logger.error("Error!")
    logger.critical("Critical!")


def test_logger():
    logger = set_logger("Loglevel_DEBUG", "DEBUG")
    _record(logger)
    logger = set_logger("Loglevel_INFO", "INFO")
    _record(logger)
    logger = set_logger("Loglevel_WARNING", "WARNING")
    _record(logger)
    logger = set_logger("Loglevel_ERROR", "ERROR")
    _record(logger)
    logger = set_logger("Loglevel_CRITICAL", "CRITICAL")
    _record(logger)

    with pytest.raises(ValueError) as excinfo:
        set_logger("Loglevel_INVALID", "INVALID")
    assert "INVALID" in str(excinfo.value)



import importlib


def set_optimizer(model, optimizer_name, **kwargs):
    """
    Set the parameter optimizer for the model.

    Reference: https://pytorch.org/docs/stable/optim.html#algorithms
    """

    torch_optim_optimizers = [
        "Adadelta",
        "Adagrad",
        "Adam",
        "AdamW",
        "Adamax",
        "ASGD",
        "RMSprop",
        "Rprop",
        "SGD",
    ]
    if optimizer_name not in torch_optim_optimizers:
        msg = "Unrecognized optimizer: {}, should be one of {}."
        raise NotImplementedError(
            msg.format(optimizer_name, ",".join(torch_optim_optimizers))
        )

    optimizer_cls = getattr(
        importlib.import_module("torch.optim"), optimizer_name
    )
    optimizer = optimizer_cls(model.parameters(), **kwargs)

    return optimizer


def update_lr(optimizer, lr):
    """
    Manually update the learning rate of the optimizer. This function is used
    when the parallelization corrupts the bindings between the optimizer and
    the scheduler.
    """

    if not lr > 0:
        msg = (
            "The learning rate should be strictly positive, but got"
            " {} instead."
        )
        raise ValueError(msg.format(lr))

    for group in optimizer.param_groups:
        group["lr"] = lr

    return optimizer


def set_scheduler(optimizer, scheduler_name, **kwargs):
    """
    Set the scheduler on learning rate for the optimizer.

    Reference:
        https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate
    """

    torch_lr_schedulers = [
        "LambdaLR",
        "MultiplicativeLR",
        "StepLR",
        "MultiStepLR",
        "ExponentialLR",
        "CosineAnnealingLR",
        "ReduceLROnPlateau",
        "CyclicLR",
        "OneCycleLR",
        "CosineAnnealingWarmRestarts",
    ]
    if scheduler_name not in torch_lr_schedulers:
        msg = "Unrecognized scheduler: {}, should be one of {}."
        raise NotImplementedError(
            msg.format(scheduler_name, ",".join(torch_lr_schedulers))
        )

    scheduler_cls = getattr(
        importlib.import_module("torch.optim.lr_scheduler"), scheduler_name
    )
    scheduler = scheduler_cls(optimizer, **kwargs)

    return scheduler



from typing import Iterable

import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
from matplotlib.ticker import MaxNLocator

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.hints import *
from rrt_ml.utilities.infos import *
from rrt_ml.utilities.paths import *


class Node:
    """
    Node (or pose) for the RRT tree.
    """

    def __init__(
            self,
            pose: Vector3,
            parent: None or 'Node' = None,
            cost_from_parent: None or 'Node' = None,
            cost_from_root: None | float = None,
            origin: None | str = None,
    ):
        """
        Initialize.
        :param pose: vector [x y theta]
        :param parent: parent pose
        :param cost_from_parent: cost of moving from parent to this pose
        :param cost_from_root: cost of moving from root node to this pose
        :param origin: whether the node comes from sampler, random, etc
        """

        self.pose = pose  # type: Vector3
        self.parent = parent  # type: None | Node
        self.cost_from_parent = cost_from_parent  # type: None | float
        self.cost_from_root = cost_from_root  # type: None | float
        self.origin = origin  # type: None | str

        self.children = []  # type: list[Node, ...]
        self.info = None  # type: None | NodeReachInfo

        self._draw_cost_from_root = False
        self._draw_cost_from_parent = False
        self._draw_color = 'black'
        self._draw_info = False
        self._draw_temp_distance = False
        self._draw_temp_info = False
        self._draw_temp_cost_from_root = False
        self._mpl_txts = []
        self._mpl_patches = []
        self._mpl_lines = []
        self._temp_distance = None
        self._temp_info = None
        self._temp_cost_from_root = None

    def set_cost_change_propagation(self):
        """
        When the 'cost_from_root' changes we need to update the 'cost_from_root' of all children.
        """

        for node_child in self.children:
            node_child.cost_from_root = (
                    self.cost_from_root + node_child.cost_from_parent
            )
            node_child.set_cost_change_propagation()

    def set_remove_me_as_child(self):
        """
        This node may belong to a list of children, and we may want to remove it.
        """

        if self.parent is not None:
            self.parent.children.remove(self)

    def set_info(self, node_reach_info: 'NodeReachInfo'):
        """
        Add stats on how this node was reached from parent node.
        """

        # Get copy of bullet stats
        self.info = node_reach_info

    def __repr__(self):
        """
        Print node.
        """

        if self.parent is not None:
            return (
                f"Pose: {self.pose}\n"
                f"Parent: {self.parent.pose}\n"
                f"Cost from parent: {self.cost_from_parent}\n"
                f"Cost from root: {self.cost_from_root}\n\n"
            )
        else:
            return f"Pose: {self.pose}\n" \
                   f"Cost from root: {self.cost_from_root}\n\n"


class Map:

    def __init__(self, cfg: MasterConfig):

        self.cfg = cfg
        self.temp_arrows = []
        self.temp_lines = []
        self.temp_texts = []
        self.nodes = []
        self.vertices = []
        self.orientations = []

        self._setup()

    def set_add_nodes(self, nodes: Iterable['Node'] | 'Node'):
        """
        Add node(s) to list of nodes.
        :param nodes: node(s) to add.
        """

        if isinstance(nodes, Iterable):
            for node in nodes:
                self.nodes.append(node)
        else:
            self.nodes.append(nodes)

    def set_add_states(self, states: Iterable[Vector4] | Vector4, origin: str = 'sl'):
        """
        Add states to list of nodes
        :param states: list of states to add
        :param origin: where the states come from
        """

        # Handle single state case
        if isinstance(states[0], Iterable):
            pass
        else:
            states = [states]

        # Iterate and add to map's list of nodes
        for st in states:
            pose = [st[0], st[1], np.arctan2(st[2], st[3])]
            self.nodes.append(Node(pose, None, None, None, origin))

    def get_plot_tree(self, ax: plt.Axes | None = None) -> tuple[plt.Figure, plt.Axes] | plt.Axes:
        """
        Plot map with obstacles and nodes/connections as a tree.
        :return: figure and axes.
        """

        # Handle cases where axis is not given and must return a new figure
        fig = None
        if ax is None:
            fig, ax = plt.subplots()

        # Set up axis layout
        self._set_plot_layout(ax)

        # Add obstacles
        self._set_plot_obstacles(ax)

        # Add branches
        self._set_plot_branches(ax)
        
        if fig is None:
            return ax
        
        else:
            return fig, ax

    def get_plot_lines(self, ax: plt.Axes | None = None) -> tuple[plt.Figure, plt.Axes] | plt.Axes:
        """
        Plot map with nodes and line connections.
        :return: figure and axes or axes
        """

        # Handle cases where axis is not given and must return a new figure
        fig = None
        if ax is None:
            fig, ax = plt.subplots()

        # Set up axis layout
        self._set_plot_layout(ax)

        # Add obstacles
        self._set_plot_obstacles(ax)

        # Add nodes
        self._set_plot_nodes(ax)

        # Add connections
        self._set_plot_connections_lines(ax)
        
        if fig is None:
            return ax

        else:
            return fig, ax

    def set_add_temp_line(self, ax: plt.Axes, node: Node, color: str):
        """
        Add temporary node to the tree (for debugging)
        """

        self._setup_ignore_static_warnings()

        # Plot line
        line = ax.plot(node.info.xs, node.info.ys, '--', color=color)

        # Add arrow to list of temp arrows
        self.temp_lines.append(line)

    def set_add_temp_arrows(self, ax: plt.Axes, nodes: Node | list[Node, ...], color: str):
        """
        Add temporary node to the tree (for debugging)
        """

        self._setup_ignore_static_warnings()
        if not isinstance(nodes, Iterable):
            nodes = [nodes]

        # Set params for different types of nodes
        kw = dict(width=0.01, head_width=0.15, head_length=0.15, ec=color, fc=color)

        for node in nodes:
            # Get arrow params
            x, y = node.pose[0:2]
            dx, dy = 0.45 * np.cos(node.pose[2]), 0.45 * np.sin(node.pose[2])

            # Plot arrow
            arrow = ax.arrow(x, y, dx, dy, **kw)

            # Add arrow to list of temp arrows
            self.temp_arrows.append(arrow)

    def set_add_temp_text(self, ax: plt.Axes, pose, text, color):
        """
        Add temporary text to the tree (for debugging)
        """

        self._setup_ignore_static_warnings()

        # Plot text
        xp, yp = pose[:2]
        # xt, yt = pose[0] + 1, pose[1] + 1
        txt = ax.annotate(text=text, xy=(xp, yp), xycoords='data', color=color)
        # txt = ax.annotate(text=text, xy=(xp, yp), xycoords='data', xytext=(xt, yt), textcoords='data')

        # Add arrow to list of temp arrows
        self.temp_texts.append(txt)

    def set_clear_temp_objs(self):
        """
        Clear matplotlib temporary objects.
        """

        for line in self.temp_lines:
            line[0].remove()
        for arrow in self.temp_arrows:
            arrow.remove()
        for text in self.temp_texts:
            text.remove()
        self.temp_lines, self.temp_arrows, self.temp_texts = [], [], []

    def _set_plot_branches(self, ax: plt.Axes):
        """
        Add nodes and connections as branches.
        :param ax: axes to plot
        """

        # Plot connection for all nodes
        for node in self.nodes:  # type: Node

            # Initial node has no parent
            if node.parent is not None:

                # Now we take (x, y) pairs, i.e. the trajectory, from stats
                xs = node.info.xs
                ys = node.info.ys

                # Plot on ax
                ax.plot(xs, ys, color='grey', alpha=0.5)

    def _set_plot_branches_for_list(self, ax: plt.Axes, list_nodes: list['Node', ...]):
        """
        Plot branches for a list of nodes only.
        :param list_nodes: list of nodes to plot branches.
        """

        # Ignore warnings
        self._setup_ignore_static_warnings()

        # Loop provided nodes
        for node in list_nodes:
            
            # Initial node has no parent
            if node.parent is not None:

                # Now we take (x, y) pairs, i.e. the trajectory, from stats
                xs = node.info.xs
                ys = node.info.ys

                # Plot on ax
                ax.plot(xs, ys, color='blue')

    def _set_plot_connections_lines(self, ax: plt.Axes):
        """
        Add connections between nodes.
        :param ax: axes do draw connections
        """

        for node in self.nodes:

            # If node has no parent continue
            if node.parent is None:
                continue

            # Get positions
            x_i, y_i, orn_i = node.pose
            x_f, y_f, orn_f = node.parent.pose

            # Add a line from root to destination point
            ax.plot([x_i, x_f], [y_i, y_f], color="k", linewidth=1.5, alpha=0.2)

    def _set_plot_layout(self, ax: plt.Axes):
        """
        Adjust aspect ratio, grid, etc.
        :param ax: axis to change.
        """

        # Make limits according to map size
        match self.cfg.maps.general.map_name:
            case 'narrow':
                # self.cfg.maps.narrow.size
                ax.set_xlim(-0.5, 0.5 + self.cfg.maps.narrow.size)
                ax.set_ylim(-0.5, 0.5 + self.cfg.maps.narrow.size)
            case _:
                pass

        # Keep figure height and width equal
        ax.set_aspect("equal", adjustable="box")

        # Tick integer values only
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

        self._setup_ignore_static_warnings()

        return ax

    def _set_plot_nodes(self, ax: plt.Axes, nodes: list[Node, ...] | None = None):
        """
        Add nodes to plot.
        :param ax: axes to plot.
        """

        # Set list of nodes to plot
        if nodes is None:
            nodes = self.nodes

        # Set params for different types of nodes
        kw_init = dict(width=0.03, head_width=0.3, head_length=0.3, ec="g", fc="g", alpha=1)
        kw_final = dict(width=0.03, head_width=0.3, head_length=0.3, ec="r", fc="r", alpha=1)
        kw_final_new = dict(width=0.01, head_width=0.15, head_length=0.15, ec="r", fc="r")
        kw_sl = dict(width=0.01, head_width=0.15, head_length=0.15, ec="k", fc="k")
        kw_random = dict(width=0.01, head_width=0.15, head_length=0.15, ec="k", fc="k")
        origin_kwargs = dict(init=kw_init, final=kw_final, final_new=kw_final_new, sl=kw_sl, random=kw_random)

        # Iterate over nodes
        for node in nodes:

            # Get arrow params
            x, y = node.pose[0:2]
            dx, dy = 0.45*np.cos(node.pose[2]), 0.45*np.sin(node.pose[2])

            # Set higher z-order for init and final node
            zorder = 3 if node.origin == 'final' or node.origin == 'init' else 1

            # Plot arrow
            ax.arrow(x, y, dx, dy, **origin_kwargs[node.origin], zorder=zorder)

    def _set_plot_obstacles(self, ax: plt.Axes):
        """
        Get rectangles patches.
        :param ax: axes to plot.
        """

        patches = []
        for v, o in zip(self.vertices, self.orientations):
            # Get params
            width = v[2] - v[0]
            height = v[1] - v[3]
            x = v[0]
            y = v[1] - height

            # Create patch and add to list
            rect = Rectangle((x, y), width, height, o)
            patches.append(rect)

        # Create black patch collection
        patches = PatchCollection(patches)
        patches.set_color("black")

        # Add to plot
        ax.add_collection(patches)

    def _set_wall(self, top_left_bottom_right: Vector4, space_snwe: Vector4 = (0, 0, 0, 0)):
        """
        Set a wall to map as a rectangle, making it shorter to not collide with other walls
        :param top_left_bottom_right: top left and bottom right vertices positions
        :param space_snwe: cut this much from each side of the rectangle (south-north-west-east)
        """

        # TopLeft and BottomRight aliases
        tl = top_left_bottom_right[:2]
        br = top_left_bottom_right[2:]

        # SpaceNorth, SpaceSouth, SpaceWest and SpaceEast aliases
        ss, sn, sw, se = space_snwe

        # Move top left vertex
        tl = [tl[0] + sw, tl[1] - sn]

        # Move bottom right vertex
        br = [br[0] - se, br[1] + ss]

        self.vertices.append([*tl, *br])
        self.orientations.append(0)

    def _setup(self):
        """
        Set up.
        """

        match self.cfg.maps.general.map_name:
            case 'arena':
                self._setup_arena()
            case 'narrow':
                self._setup_narrow()
            case _:
                raise NotImplementedError

    def _setup_arena(self, length=None, width=None, offset=None):
        """
        Set a rectangular arena.
        :param length: length of arena
        :param width: width of arena
        :param offset: offset center of arena
        """

        # Config constants
        if length is None:
            length = self.cfg.maps.arena.length
        if width is None:
            width = self.cfg.maps.arena.width
        if offset is None:
            offset = self.cfg.maps.arena.offset
        ox, oy = offset

        # Fine tune constants: WallThickness, Spacing
        wt = 0.3
        s = 1e-3

        # Make vertices
        left_wall_pos = [-wt + ox, width + oy, ox, oy]
        right_wall_pos = [length + ox, width + oy, length + wt + ox, oy]
        bottom_wall_pos = [ox, oy, length + ox, -wt + oy]
        top_wall_pos = [ox, width + wt + oy, length + ox, width]

        # Set walls with spacing
        self._set_wall(bottom_wall_pos, space_snwe=[0, s, 0, 0])
        self._set_wall(top_wall_pos, space_snwe=[s, 0, 0, 0])
        self._set_wall(left_wall_pos, space_snwe=[0, 0, 0, s])
        self._set_wall(right_wall_pos, space_snwe=[0, 0, s, 0])

    def _setup_narrow(self):
        """
        Set up narrow map.
        """

        # Config constants
        size = self.cfg.maps.narrow.size
        x1, y1 = self.cfg.maps.narrow.narrow1_pos
        x2, y2 = self.cfg.maps.narrow.narrow2_pos
        ww = self.cfg.maps.narrow.wall_width
        pw = self.cfg.maps.narrow.passage_width

        # Fine tune constants: Spacing
        s = 1e-3

        # Box around map
        self._setup_arena(size, size, offset=(0, 0))

        # Left lower, left upper, right lower and right upper walls
        ll = [x1 - ww / 2, y1 - pw / 2, x1 + ww / 2, 0]
        lu = [x1 - ww / 2, size, x1 + ww / 2, y1 + pw / 2]
        rl = [x2 - ww / 2, y2 - pw / 2, x2 + ww / 2, 0]
        ru = [x2 - ww / 2, size, x2 + ww / 2, y2 + pw / 2]

        # Set walls
        self._set_wall(ll, space_snwe=[s, 0, 0, 0])
        self._set_wall(lu, space_snwe=[0, s, 0, 0])
        self._set_wall(rl, space_snwe=[s, 0, 0, 0])
        self._set_wall(ru, space_snwe=[0, s, 0, 0])

    def _setup_ignore_static_warnings(self):
        """
        Ignore static warnings.
        """
        pass




from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *


# Set stats to true to track bullet stats
cfg = MasterConfig()
cfg.env.general.gui = True

# Reset master to track stats
env = CarNavigationBulletEnv(cfg)
env.target.set_pose([0, 0, 0])

env.car.set_pose([0.0, 0.0, 0])
obs = env._get_observation()
print(env.compute_reward(obs['achieved_goal'], obs['desired_goal']))
img = env._get_image_current_top_view()
img.save('reward_episilon_good_1.png')

env.car.set_pose([0.01, 0.01, 0])
obs = env._get_observation()
print(env.compute_reward(obs['achieved_goal'], obs['desired_goal']))
img = env._get_image_current_top_view()
img.save('reward_episilon_good_2.png')

env.car.set_pose([0, 0, 0.002])
obs = env._get_observation()
print(env.compute_reward(obs['achieved_goal'], obs['desired_goal']))
img = env._get_image_current_top_view()
img.save('reward_episilon_good_3.png')

env.car.set_pose([0.02, 0.01, 0])
obs = env._get_observation()
print(env.compute_reward(obs['achieved_goal'], obs['desired_goal']))
img = env._get_image_current_top_view()
img.save('reward_episilon_bad_1.png')

env.car.set_pose([0.01, 0.02, 0.04])
obs = env._get_observation()
print(env.compute_reward(obs['achieved_goal'], obs['desired_goal']))
img = env._get_image_current_top_view()
img.save('reward_episilon_bad_2.png')

env.car.set_pose([0.01, 0.01, 0.06])
obs = env._get_observation()
print(env.compute_reward(obs['achieved_goal'], obs['desired_goal']))
img = env._get_image_current_top_view()
img.save('reward_episilon_bad_3.png')

pass





from mrl.import_all import *
from argparse import Namespace
import gym
import time


def make_ddpg_agent(base_config=default_ddpg_config,
                    args=Namespace(env='InvertedPendulum-v2',
                                   tb='',
                                   parent_folder='/tmp/mrl',
                                   layers=(256, 256),
                                   num_envs=None),
                    agent_name_attrs=['env', 'seed', 'tb'],
                    **kwargs):

  if callable(base_config):
    base_config = base_config()
  config = base_config

  if hasattr(args, 'num_envs') and args.num_envs is None:
    import multiprocessing as mp
    args.num_envs = max(mp.cpu_count() - 1, 1)

  if not hasattr(args, 'prefix'):
    args.prefix = 'ddpg'
  if not args.tb:
    args.tb = str(time.time())

  merge_args_into_config(args, config)

  config.agent_name = make_agent_name(config, agent_name_attrs, prefix=args.prefix)

  
  base_modules = {
      k: v
      for k, v in dict(module_train=StandardTrain(),
                       module_eval=EpisodicEval(),
                       module_policy=ActorPolicy(),
                       module_logger=Logger(),
                       module_state_normalizer=Normalizer(MeanStdNormalizer()),
                       module_replay=OnlineHERBuffer(),
                       module_action_noise=ContinuousActionNoise(GaussianProcess,
                                                                 std=ConstantSchedule(config.action_noise)),
                       module_algorithm=DDPG()).items() if not k in config
  }

  config.update(base_modules)

  if type(args.env) is str:
    env = lambda: gym.make(args.env)
    eval_env = env
  else:
    env = args.env
    eval_env = env
  
  if hasattr(args, 'eval_env') and args.eval_env is not None:
    if type(args.eval_env) is str:
      eval_env = lambda: gym.make(args.eval_env)
    else:
      eval_env = args.eval_env


    
  config.module_train_env = EnvModule(env, num_envs=config.num_envs, seed=config.seed)
  config.module_eval_env = EnvModule(eval_env, num_envs=config.num_eval_envs, name='eval_env', seed=config.seed + 1138)

  layer_norm = nn.LayerNorm if (hasattr(args, 'layer_norm') and args.layer_norm) else nn.Identity

  e = config.module_eval_env
  config.module_actor = PytorchModel(
      'actor', lambda: Actor(FCBody(e.state_dim + e.goal_dim, args.layers, layer_norm, make_activ(config.activ)), e.action_dim, e.max_action))
  config.module_critic = PytorchModel(
      'critic', lambda: Critic(FCBody(e.state_dim + e.goal_dim + e.action_dim, args.layers, layer_norm, make_activ(config.activ)), 1))

  if e.goal_env:
    config.never_done = True # important for standard Gym goal environments, which are never done

  return config


def make_td3_agent(base_config=spinning_up_td3_config,
                   args=Namespace(env='InvertedPendulum-v2',
                                  tb='',
                                  prefix='td3',
                                  parent_folder='/tmp/mrl',
                                  layers=(256, 256),
                                  num_envs=None),
                   agent_name_attrs=['env', 'seed', 'tb'],
                   **kwargs):
  
  config = make_ddpg_agent(base_config, args, agent_name_attrs, **kwargs)
  del config.module_algorithm
  config.module_algorithm = TD3()

  layer_norm = nn.LayerNorm if (hasattr(args, 'layer_norm') and args.layer_norm) else nn.Identity
  
  e = config.module_eval_env
  config.module_critic2 = PytorchModel('critic2',
      lambda: Critic(FCBody(e.state_dim + e.goal_dim + e.action_dim, args.layers, layer_norm, make_activ(config.activ), False), 1, False))

  return config


def make_sac_agent(base_config=spinning_up_sac_config,
                   args=Namespace(env='InvertedPendulum-v2',
                                  tb='',
                                  prefix='sac',
                                  parent_folder='/tmp/mrl',
                                  layers=(256, 256),
                                  num_envs=None),
                   agent_name_attrs=['env', 'seed', 'tb'],
                   **kwargs):
  
  config = make_ddpg_agent(base_config, args, agent_name_attrs, **kwargs)
  e = config.module_eval_env
  layer_norm = nn.LayerNorm if (hasattr(args, 'layer_norm') and args.layer_norm) else nn.Identity
  
  del config.module_actor
  del config.module_action_noise
  del config.module_policy
  config.module_policy = StochasticActorPolicy()
  del config.module_algorithm
  config.module_algorithm = SAC()

  config.module_actor = PytorchModel(
      'actor', lambda: StochasticActor(FCBody(e.state_dim + e.goal_dim, args.layers, layer_norm, make_activ(config.activ)), 
        e.action_dim, e.max_action, log_std_bounds = (-20, 2)))

  config.module_critic2 = PytorchModel('critic2',
      lambda: Critic(FCBody(e.state_dim + e.goal_dim + e.action_dim, args.layers, layer_norm, make_activ(config.activ), False), 1, False))

  return config


import matplotlib.pyplot as plt
import numpy as np

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.paths import *


cfg = MasterConfig()

path = Paths().experiments / 'others' / 'alpha.png'

i = np.arange(1, 1000)
a, b, c = cfg.rrt.sample.learner_prob
sl_prob = a*np.exp(-(i - b)**2/(2*c**2))

fig, ax = plt.subplots()
ax.plot(i, sl_prob)
ax.grid(True)
ax.set_xlabel('Iteração')
ax.set_ylabel(r'$\alpha$')
ax.set_xlim([0, 500])

fig.savefig(path, dpi=600, bbox_inches='tight')



__model_doc = """
    Parameters
    ----------
    estimator : torch.nn.Module
        The class or object of your base estimator.

        - If :obj:`class`, it should inherit from :mod:`torch.nn.Module`.
        - If :obj:`object`, it should be instantiated from a class inherited
          from :mod:`torch.nn.Module`.
    n_estimators : int
        The number of base estimators in the ensemble.
    estimator_args : dict, default=None
        The dictionary of hyper-parameters used to instantiate base
        estimators. This parameter will have no effect if ``estimator`` is a
        base estimator object after instantiation.
    cuda : bool, default=True

        - If ``True``, use GPU to train and evaluate the ensemble.
        - If ``False``, use CPU to train and evaluate the ensemble.
    n_jobs : int, default=None
        The number of workers for training the ensemble. This input
        argument is used for parallel ensemble methods such as
        :mod:`voting` and :mod:`bagging`. Setting it to an integer larger
        than ``1`` enables ``n_jobs`` base estimators to be trained
        simultaneously.

    Attributes
    ----------
    estimators_ : torch.nn.ModuleList
        An internal container that stores all fitted base estimators.
"""


__seq_model_doc = """
    Parameters
    ----------
    estimator : torch.nn.Module
        The class or object of your base estimator.

        - If :obj:`class`, it should inherit from :mod:`torch.nn.Module`.
        - If :obj:`object`, it should be instantiated from a class inherited
          from :mod:`torch.nn.Module`.
    n_estimators : int
        The number of base estimators in the ensemble.
    estimator_args : dict, default=None
        The dictionary of hyper-parameters used to instantiate base
        estimators. This parameter will have no effect if ``estimator`` is a
        base estimator object after instantiation.
    cuda : bool, default=True

        - If ``True``, use GPU to train and evaluate the ensemble.
        - If ``False``, use CPU to train and evaluate the ensemble.

    Attributes
    ----------
    estimators_ : torch.nn.ModuleList
        An internal container that stores all fitted base estimators.
"""


__set_optimizer_doc = """
    Parameters
    ----------
    optimizer_name : string
        The name of the optimizer, should be one of {``Adadelta``, ``Adagrad``,
        ``Adam``, ``AdamW``, ``Adamax``, ``ASGD``, ``RMSprop``, ``Rprop``,
        ``SGD``}.
    **kwargs : keyword arguments
        Keyword arguments on setting the optimizer, should be in the form:
        ``lr=1e-3, weight_decay=5e-4, ...``. These keyword arguments
        will be directly passed to :mod:`torch.optim.Optimizer`.
"""


__set_scheduler_doc = """
    Parameters
    ----------
    scheduler_name : string
        The name of the scheduler, should be one of {``LambdaLR``,
        ``MultiplicativeLR``, ``StepLR``, ``MultiStepLR``, ``ExponentialLR``,
        ``CosineAnnealingLR``, ``ReduceLROnPlateau``}.
    **kwargs : keyword arguments
        Keyword arguments on setting the scheduler. These keyword arguments
        will be directly passed to :mod:`torch.optim.lr_scheduler`.
"""


__set_criterion_doc = """
    Parameters
    ----------
    criterion : torch.nn.loss
        The customized training criterion object.
"""


__fit_doc = """
    Parameters
    ----------
    train_loader : torch.utils.data.DataLoader
        A :mod:`torch.utils.data.DataLoader` container that contains the
        training data.
    epochs : int, default=100
        The number of training epochs.
    log_interval : int, default=100
        The number of batches to wait before logging the training status.
    test_loader : torch.utils.data.DataLoader, default=None
        A :mod:`torch.utils.data.DataLoader` container that contains the
        evaluating data.

        - If ``None``, no validation is conducted during the training
          stage.
        - If not ``None``, the ensemble will be evaluated on this
          dataloader after each training epoch.
    save_model : bool, default=True
        Specify whether to save the model parameters.

        - If test_loader is ``None``, the ensemble fully trained will be
          saved.
        - If test_loader is not ``None``, the ensemble with the best
          validation performance will be saved.
    save_dir : string, default=None
        Specify where to save the model parameters.

        - If ``None``, the model will be saved in the current directory.
        - If not ``None``, the model will be saved in the specified
          directory: ``save_dir``.
"""


__predict_doc = """
    Return the predictions of the ensemble given the testing data.

    Parameters
    ----------
    X : {tensor, numpy array}
        A data batch in the form of tensor or numpy array.

    Returns
    -------
    pred : tensor of shape (n_samples, n_outputs)
        For classifiers, ``n_outputs`` is the number of distinct classes. For
        regressors, ``n_output`` is the number of target variables.
"""


__classification_forward_doc = """
    Parameters
    ----------
    X : tensor
        An input batch of data, which should be a valid input data batch
        for base estimators in the ensemble.

    Returns
    -------
    proba : tensor of shape (batch_size, n_classes)
        The predicted class distribution.
"""


__classification_evaluate_doc = """
    Compute the classification accuracy of the ensemble given the testing
    dataloader and optionally the average cross-entropy loss.

    Parameters
    ----------
    test_loader : torch.utils.data.DataLoader
        A data loader that contains the testing data.
    return_loss : bool, default=False
        Whether to return the average cross-entropy loss over all batches
        in the ``test_loader``.

    Returns
    -------
    accuracy : float
        The classification accuracy of the fitted ensemble on ``test_loader``.
    loss : float
        The average cross-entropy loss of the fitted ensemble on
        ``test_loader``, only available when ``return_loss`` is True.
"""


__regression_forward_doc = """
    Parameters
    ----------
    X : tensor
        An input batch of data, which should be a valid input data batch
        for base estimators in the ensemble.

    Returns
    -------
    pred : tensor of shape (batch_size, n_outputs)
        The predicted values.
"""


__regression_evaluate_doc = """
    Compute the mean squared error (MSE) of the ensemble given the testing
    dataloader.

    Parameters
    ----------
    test_loader : torch.utils.data.DataLoader
        A data loader that contains the testing data.

    Returns
    -------
    mse : float
        The testing mean squared error of the fitted ensemble on
        ``test_loader``.
"""



import torch
import pytest
import torchensemble
import torch.nn as nn


# Base estimator
class MLP(nn.Module):
    def __init__(self):
        super(MLP, self).__init__()
        self.linear1 = nn.Linear(2, 2)
        self.linear2 = nn.Linear(2, 2)

    def forward(self, X):
        X = X.view(X.size()[0], -1)
        output = self.linear1(X)
        output = self.linear2(output)
        return output


def test_set_scheduler_LambdaLR():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    lr_lambda = lambda x: x * 0.1  # noqa: E731
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "LambdaLR", lr_lambda=lr_lambda
    )


def test_set_scheduler_MultiplicativeLR():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    lr_lambda = lambda x: x * 0.1  # noqa: E731
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "MultiplicativeLR", lr_lambda=lr_lambda
    )


def test_set_scheduler_StepLR():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "StepLR", step_size=50
    )


def test_set_scheduler_MultiStepLR():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "MultiStepLR", milestones=[50, 100]
    )


def test_set_scheduler_ExponentialLR():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "ExponentialLR", gamma=0.1
    )


def test_set_scheduler_CosineAnnealingLR():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "CosineAnnealingLR", T_max=100
    )


def test_set_scheduler_ReduceLROnPlateau():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())
    torchensemble.utils.set_module.set_scheduler(
        optimizer, "ReduceLROnPlateau"
    )


def test_set_scheduler_Unknown():
    model = MLP()
    optimizer = torch.optim.Adam(model.parameters())

    with pytest.raises(NotImplementedError) as excinfo:
        torchensemble.utils.set_module.set_scheduler(optimizer, "Unknown")
    assert "Unrecognized scheduler" in str(excinfo.value)



from multiprocessing import Process, Pipe

import numpy as np

from mrl.utils.vec_env import VecEnv, CloudpickleWrapper
from gym import spaces


def _worker(remote, parent_remote, env_fn_wrapper):
  parent_remote.close()
  env = env_fn_wrapper.var()
  while True:
    try:
      cmd, data = remote.recv()
      if cmd == 'step':
        observation, reward, done, info = env.step(data)
        if done:
          observation = env.reset()
        remote.send((observation, reward, done, info))
      elif cmd == 'reset':
        observation = env.reset()
        remote.send(observation)
      elif cmd == 'render':
        remote.send(env.render(*data[0], **data[1]))
      elif cmd == 'close':
        remote.close()
        break
      elif cmd == 'get_spaces':
        remote.send((env.observation_space, env.action_space))
      elif cmd == 'env_method':
        method = getattr(env, data[0])
        remote.send(method(*data[1], **data[2]))
      elif cmd == 'get_attr':
        remote.send(getattr(env, data))
      elif cmd == 'set_attr':
        remote.send(setattr(env, data[0], data[1]))
      else:
        raise NotImplementedError
    except EOFError:
      break


class SubprocVecEnv(VecEnv):
  """
    Creates a multiprocess vectorized wrapper for multiple environments

    :param env_fns: ([Gym Environment]) Environments to run in subprocesses
    """

  def __init__(self, env_fns):
    self.waiting = False
    self.closed = False
    n_envs = len(env_fns)

    self.remotes, self.work_remotes = zip(*[Pipe() for _ in range(n_envs)])
    self.processes = [
        Process(target=_worker, args=(work_remote, remote, CloudpickleWrapper(env_fn)))
        for (work_remote, remote, env_fn) in zip(self.work_remotes, self.remotes, env_fns)
    ]
    for process in self.processes:
      process.daemon = True  # if the main process crashes, we should not cause things to hang
      process.start()
    for remote in self.work_remotes:
      remote.close()

    self.remotes[0].send(('get_spaces', None))
    observation_space, action_space = self.remotes[0].recv()

    self.goal_env = False
    self.goal_keys = None
    if isinstance(observation_space, spaces.Dict):
      dummy_env = env_fns[0]()
      self.dummy_env = dummy_env
      if dummy_env.compute_reward is not None:
        self.compute_reward = dummy_env.compute_reward

      if hasattr(dummy_env, 'goal_extraction_function') and dummy_env.goal_extraction_function is not None:
        self.goal_extraction_function = dummy_env.goal_extraction_function
      self.goal_env = True
      self.goal_keys = tuple(observation_space.spaces.keys())
    VecEnv.__init__(self, len(env_fns), observation_space, action_space)

  def step_async(self, actions):
    for remote, action in zip(self.remotes, actions):
      remote.send(('step', action))
    self.waiting = True

  def step_wait(self):
    results = [remote.recv() for remote in self.remotes]
    self.waiting = False
    obs, rews, dones, infos = zip(*results)
    if self.goal_env:
      obs = {k: np.stack([o[k] for o in obs]) for k in self.goal_keys}
    else:
      obs = np.stack(obs)
    return obs, np.stack(rews), np.stack(dones), infos

  def reset(self):
    for remote in self.remotes:
      remote.send(('reset', None))
    obs = [remote.recv() for remote in self.remotes]
    if self.goal_env:
      obs = {k: np.stack([o[k] for o in obs]) for k in self.goal_keys}
    else:
      obs = np.stack(obs)
    return obs

  def close(self):
    if self.closed:
      return
    if self.waiting:
      for remote in self.remotes:
        remote.recv()
    for remote in self.remotes:
      remote.send(('close', None))
    for process in self.processes:
      process.join()
    self.closed = True

  def render(self, mode='human', *args, **kwargs):
    for pipe in self.remotes:
      # gather images from subprocesses
      # `mode` will be taken into account later
      pipe.send(('render', (args, {'mode': 'rgb_array', **kwargs})))
    imgs = [pipe.recv() for pipe in self.remotes]
    # Create a big image by tiling images from subprocesses
    bigimg = tile_images(imgs)
    if mode == 'human':
      import cv2
      cv2.imshow('vecenv', bigimg[:, :, ::-1])
      cv2.waitKey(1)
    elif mode == 'rgb_array':
      return bigimg
    else:
      raise NotImplementedError

  def get_images(self):
    for pipe in self.remotes:
      pipe.send(('render', {"mode": 'rgb_array'}))
    imgs = [pipe.recv() for pipe in self.remotes]
    return imgs

  def get_attr(self, attr_name, indices=None):
      """Return attribute from vectorized environment (see base class)."""
      target_remotes = self._get_target_remotes(indices)
      for remote in target_remotes:
          remote.send(('get_attr', attr_name))
      return [remote.recv() for remote in target_remotes]

  def set_attr(self, attr_name, value, indices=None):
      """Set attribute inside vectorized environments (see base class)."""
      target_remotes = self._get_target_remotes(indices)
      for remote in target_remotes:
          remote.send(('set_attr', (attr_name, value)))
      for remote in target_remotes:
          remote.recv()

  def env_method(self, method_name, *method_args, indices=None, **method_kwargs):
      """Call instance methods of vectorized environments."""
      target_remotes = self._get_target_remotes(indices)
      for remote in target_remotes:
          remote.send(('env_method', (method_name, method_args, method_kwargs)))
      return [remote.recv() for remote in target_remotes]

  def _get_target_remotes(self, indices):
      """
      Get the connection object needed to communicate with the wanted
      envs that are in subprocesses.
      :param indices: (None,int,Iterable) refers to indices of envs.
      :return: ([multiprocessing.Connection]) Connection object to communicate between processes.
      """
      indices = self._get_indices(indices)
      return [self.remotes[i] for i in indices]



def tile_images(img_nhwc):
  """
    Tile N images into one big PxQ image
    (P,Q) are chosen to be as close as possible, and if N
    is square, then P=Q.

    :param img_nhwc: (list) list or array of images, ndim=4 once turned into array. img nhwc
        n = batch index, h = height, w = width, c = channel
    :return: (numpy float) img_HWc, ndim=3
    """
  img_nhwc = np.asarray(img_nhwc)
  n_images, height, width, n_channels = img_nhwc.shape
  # new_height was named H before
  new_height = int(np.ceil(np.sqrt(n_images)))
  # new_width was named W before
  new_width = int(np.ceil(float(n_images) / new_height))
  img_nhwc = np.array(list(img_nhwc) + [img_nhwc[0] * 0 for _ in range(n_images, new_height * new_width)])
  # img_HWhwc
  out_image = img_nhwc.reshape(new_height, new_width, height, width, n_channels)
  # img_HhWwc
  out_image = out_image.transpose(0, 2, 1, 3, 4)
  # img_Hh_Ww_c
  out_image = out_image.reshape(new_height * height, new_width * width, n_channels)
  return out_image



import torch
import pytest
import numpy as np
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

from torchensemble.utils import operator as op


outputs = [
    torch.FloatTensor(np.array(([1, 1], [1, 1]))),
    torch.FloatTensor(np.array(([2, 2], [2, 2]))),
    torch.FloatTensor(np.array(([3, 3], [3, 3]))),
]

label = torch.LongTensor(np.array(([0, 1, 2, 1])))
n_classes = 3


def test_average():
    actual = op.average(outputs).numpy()
    expected = np.array(([2, 2], [2, 2]))
    assert_array_equal(actual, expected)


def test_sum_with_multiplicative():
    shrinkage_rate = 0.1
    actual = op.sum_with_multiplicative(outputs, shrinkage_rate).numpy()
    expected = np.array(([0.6, 0.6], [0.6, 0.6]))
    assert_array_almost_equal(actual, expected)


def test_onehot_encoding():
    actual = op.onehot_encoding(label, n_classes).numpy()
    expected = np.array(([1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 0]))
    assert_array_almost_equal(actual, expected)


def test_residual_regression_invalid_shape():
    with pytest.raises(ValueError) as excinfo:
        op.pseudo_residual_regression(
            torch.FloatTensor(np.array(([1, 1], [1, 1]))),  # 2 * 2
            label.view(-1, 1),  # 4 * 1
        )
    assert "should be the same as output" in str(excinfo.value)



"""
  Snapshot ensemble generates many base estimators by enforcing a base
  estimator to converge to its local minima many times and save the
  model parameters at that point as a snapshot. The final prediction takes
  the average over predictions from all snapshot models.

  Reference:
      G. Huang, Y.-X. Li, G. Pleiss et al., Snapshot Ensemble: Train 1, and
      M for free, ICLR, 2017.
"""


import math
import torch
import logging
import warnings
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import LambdaLR

from ._base import BaseModule, BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op
from .utils.logging import get_tb_logger


__all__ = ["SnapshotEnsembleClassifier", "SnapshotEnsembleRegressor"]


__fit_doc = """
    Parameters
    ----------
    train_loader : torch.utils.data.DataLoader
        A :mod:`DataLoader` container that contains the training data.
    lr_clip : list or tuple, default=None
        Specify the accepted range of learning rate. When the learning rate
        determined by the scheduler is out of this range, it will be clipped.

        - The first element should be the lower bound of learning rate.
        - The second element should be the upper bound of learning rate.
    epochs : int, default=100
        The number of training epochs.
    log_interval : int, default=100
        The number of batches to wait before logging the training status.
    test_loader : torch.utils.data.DataLoader, default=None
        A :mod:`DataLoader` container that contains the evaluating data.

        - If ``None``, no validation is conducted after each snapshot
          being generated.
        - If not ``None``, the ensemble will be evaluated on this
          dataloader after each snapshot being generated.
    save_model : bool, default=True
        Specify whether to save the model parameters.

        - If test_loader is ``None``, the ensemble with
          ``n_estimators`` base estimators will be saved.
        - If test_loader is not ``None``, the ensemble with the best
          validation performance will be saved.
    save_dir : string, default=None
        Specify where to save the model parameters.

        - If ``None``, the model will be saved in the current directory.
        - If not ``None``, the model will be saved in the specified
          directory: ``save_dir``.
"""


def _snapshot_ensemble_model_doc(header, item="fit"):
    """
    Decorator on obtaining documentation for different snapshot ensemble
    models.
    """

    def get_doc(item):
        """Return selected item"""
        __doc = {"fit": __fit_doc}
        return __doc[item]

    def adddoc(cls):
        doc = [header + "\n\n"]
        doc.extend(get_doc(item))
        cls.__doc__ = "".join(doc)
        return cls

    return adddoc


class _BaseSnapshotEnsemble(BaseModule):
    def __init__(
        self, estimator, n_estimators, estimator_args=None, cuda=True
    ):
        super(BaseModule, self).__init__()

        self.base_estimator_ = estimator
        self.n_estimators = n_estimators
        self.estimator_args = estimator_args

        if estimator_args and not isinstance(estimator, type):
            msg = (
                "The input `estimator_args` will have no effect since"
                " `estimator` is already an object after instantiation."
            )
            warnings.warn(msg, RuntimeWarning)

        self.device = torch.device("cuda" if cuda else "cpu")
        self.logger = logging.getLogger()
        self.tb_logger = get_tb_logger()

        self.estimators_ = nn.ModuleList()

    def _validate_parameters(self, lr_clip, epochs, log_interval):
        """Validate hyper-parameters on training the ensemble."""

        if lr_clip:
            if not (isinstance(lr_clip, list) or isinstance(lr_clip, tuple)):
                msg = "lr_clip should be a list or tuple with two elements."
                self.logger.error(msg)
                raise ValueError(msg)

            if len(lr_clip) != 2:
                msg = (
                    "lr_clip should only have two elements, one for lower"
                    " bound, and another for upper bound."
                )
                self.logger.error(msg)
                raise ValueError(msg)

            if not lr_clip[0] < lr_clip[1]:
                msg = (
                    "The first element = {} should be smaller than the"
                    " second element = {} in lr_clip."
                )
                self.logger.error(msg.format(lr_clip[0], lr_clip[1]))
                raise ValueError(msg.format(lr_clip[0], lr_clip[1]))

        if not epochs > 0:
            msg = (
                "The number of training epochs = {} should be strictly"
                " positive."
            )
            self.logger.error(msg.format(epochs))
            raise ValueError(msg.format(epochs))

        if not log_interval > 0:
            msg = (
                "The number of batches to wait before printting the"
                " training status should be strictly positive, but got {}"
                " instead."
            )
            self.logger.error(msg.format(log_interval))
            raise ValueError(msg.format(log_interval))

        if not epochs % self.n_estimators == 0:
            msg = (
                "The number of training epochs = {} should be a multiple"
                " of n_estimators = {}."
            )
            self.logger.error(msg.format(epochs, self.n_estimators))
            raise ValueError(msg.format(epochs, self.n_estimators))

    def _forward(self, *x):
        """
        Implementation on the internal data forwarding in snapshot ensemble.
        """
        # Average
        results = [estimator(*x) for estimator in self.estimators_]
        output = op.average(results)

        return output

    def _clip_lr(self, optimizer, lr_clip):
        """Clip the learning rate of the optimizer according to `lr_clip`."""
        if not lr_clip:
            return optimizer

        for param_group in optimizer.param_groups:
            if param_group["lr"] < lr_clip[0]:
                param_group["lr"] = lr_clip[0]
            if param_group["lr"] > lr_clip[1]:
                param_group["lr"] = lr_clip[1]

        return optimizer

    def _set_scheduler(self, optimizer, n_iters):
        """
        Set the learning rate scheduler for snapshot ensemble.
        Please refer to the equation (2) in original paper for details.
        """
        T_M = math.ceil(n_iters / self.n_estimators)
        lr_lambda = lambda iteration: 0.5 * (  # noqa: E731
            torch.cos(torch.tensor(math.pi * (iteration % T_M) / T_M)) + 1
        )
        scheduler = LambdaLR(optimizer, lr_lambda=lr_lambda)

        return scheduler

    def set_scheduler(self, scheduler_name, **kwargs):
        msg = (
            "The learning rate scheduler for Snapshot Ensemble will be"
            " automatically set. Calling this function has no effect on"
            " the training stage of Snapshot Ensemble."
        )
        warnings.warn(msg, RuntimeWarning)


@torchensemble_model_doc(
    """Implementation on the SnapshotEnsembleClassifier.""", "seq_model"
)
class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble, BaseClassifier):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in SnapshotEnsembleClassifier.""",  # noqa: E501
        "classifier_forward",
    )
    def forward(self, *x):
        proba = self._forward(*x)

        return F.softmax(proba, dim=1)

    @torchensemble_model_doc(
        """Set the attributes on optimizer for SnapshotEnsembleClassifier.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for SnapshotEnsembleClassifier.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_snapshot_ensemble_model_doc(
        """Implementation on the training stage of SnapshotEnsembleClassifier.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        lr_clip=None,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        self._validate_parameters(lr_clip, epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        estimator = self._make_estimator()

        # Set the optimizer and scheduler
        optimizer = set_module.set_optimizer(
            estimator, self.optimizer_name, **self.optimizer_args
        )

        scheduler = self._set_scheduler(optimizer, epochs * len(train_loader))

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.CrossEntropyLoss()

        # Utils
        best_acc = 0.0
        counter = 0  # a counter on generating snapshots
        total_iters = 0
        n_iters_per_estimator = epochs * len(train_loader) // self.n_estimators

        # Training loop
        estimator.train()
        for epoch in range(epochs):
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)
                batch_size = data[0].size(0)

                # Clip the learning rate
                optimizer = self._clip_lr(optimizer, lr_clip)

                optimizer.zero_grad()
                output = estimator(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        _, predicted = torch.max(output.data, 1)
                        correct = (predicted == target).sum().item()

                        msg = (
                            "lr: {:.5f} | Epoch: {:03d} | Batch: {:03d} |"
                            " Loss: {:.5f} | Correct: {:d}/{:d}"
                        )
                        self.logger.info(
                            msg.format(
                                optimizer.param_groups[0]["lr"],
                                epoch,
                                batch_idx,
                                loss,
                                correct,
                                batch_size,
                            )
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "snapshot_ensemble/Train_Loss",
                                loss,
                                total_iters,
                            )
                        else:
                            print("None")

                # Snapshot ensemble updates the learning rate per iteration
                # instead of per epoch.
                scheduler.step()
                counter += 1
                total_iters += 1

            if counter % n_iters_per_estimator == 0:

                # Generate and save the snapshot
                snapshot = self._make_estimator()
                snapshot.load_state_dict(estimator.state_dict())
                self.estimators_.append(snapshot)

                msg = "Save the snapshot model with index: {}"
                self.logger.info(msg.format(len(self.estimators_) - 1))

            # Validation after each snapshot model being generated
            if test_loader and counter % n_iters_per_estimator == 0:
                self.eval()
                with torch.no_grad():
                    correct = 0
                    total = 0
                    for _, elem in enumerate(test_loader):
                        data, target = io.split_data_target(elem, self.device)
                        output = self.forward(*data)
                        _, predicted = torch.max(output.data, 1)
                        correct += (predicted == target).sum().item()
                        total += target.size(0)
                    acc = 100 * correct / total

                    if acc > best_acc:
                        best_acc = acc
                        if save_model:
                            io.save(self, save_dir, self.logger)

                    msg = (
                        "n_estimators: {} | Validation Acc: {:.3f} %"
                        " | Historical Best: {:.3f} %"
                    )
                    self.logger.info(
                        msg.format(len(self.estimators_), acc, best_acc)
                    )
                    if self.tb_logger:
                        self.tb_logger.add_scalar(
                            "snapshot_ensemble/Validation_Acc",
                            acc,
                            len(self.estimators_),
                        )

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@torchensemble_model_doc(
    """Implementation on the SnapshotEnsembleRegressor.""", "seq_model"
)
class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble, BaseRegressor):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in SnapshotEnsembleRegressor.""",  # noqa: E501
        "regressor_forward",
    )
    def forward(self, *x):
        pred = self._forward(*x)
        return pred

    @torchensemble_model_doc(
        """Set the attributes on optimizer for SnapshotEnsembleRegressor.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for SnapshotEnsembleRegressor.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_snapshot_ensemble_model_doc(
        """Implementation on the training stage of SnapshotEnsembleRegressor.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        lr_clip=None,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        self._validate_parameters(lr_clip, epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        estimator = self._make_estimator()

        # Set the optimizer and scheduler
        optimizer = set_module.set_optimizer(
            estimator, self.optimizer_name, **self.optimizer_args
        )

        scheduler = self._set_scheduler(optimizer, epochs * len(train_loader))

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.MSELoss()

        # Utils
        best_loss = float("inf")
        counter = 0  # a counter on generating snapshots
        total_iters = 0
        n_iters_per_estimator = epochs * len(train_loader) // self.n_estimators

        # Training loop
        estimator.train()
        for epoch in range(epochs):
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)

                # Clip the learning rate
                optimizer = self._clip_lr(optimizer, lr_clip)

                optimizer.zero_grad()
                output = estimator(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        msg = (
                            "lr: {:.5f} | Epoch: {:03d} | Batch: {:03d}"
                            " | Loss: {:.5f}"
                        )
                        self.logger.info(
                            msg.format(
                                optimizer.param_groups[0]["lr"],
                                epoch,
                                batch_idx,
                                loss,
                            )
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "snapshot_ensemble/Train_Loss",
                                loss,
                                total_iters,
                            )

                # Snapshot ensemble updates the learning rate per iteration
                # instead of per epoch.
                scheduler.step()
                counter += 1
                total_iters += 1

            if counter % n_iters_per_estimator == 0:
                # Generate and save the snapshot
                snapshot = self._make_estimator()
                snapshot.load_state_dict(estimator.state_dict())
                self.estimators_.append(snapshot)

                msg = "Save the snapshot model with index: {}"
                self.logger.info(msg.format(len(self.estimators_) - 1))

            # Validation after each snapshot model being generated
            if test_loader and counter % n_iters_per_estimator == 0:
                self.eval()
                with torch.no_grad():
                    val_loss = 0.0
                    for _, elem in enumerate(test_loader):
                        data, target = io.split_data_target(elem, self.device)
                        output = self.forward(*data)
                        val_loss += self._criterion(output, target)
                    val_loss /= len(test_loader)

                    if val_loss < best_loss:
                        best_loss = val_loss
                        if save_model:
                            io.save(self, save_dir, self.logger)

                    msg = (
                        "n_estimators: {} | Validation Loss: {:.5f} |"
                        " Historical Best: {:.5f}"
                    )
                    self.logger.info(
                        msg.format(len(self.estimators_), val_loss, best_loss)
                    )
                    if self.tb_logger:
                        self.tb_logger.add_scalar(
                            "snapshot_ensemble/Validation_Loss",
                            val_loss,
                            len(self.estimators_),
                        )

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



# flake8: noqa F401
from mrl.utils.vec_env.base_vec_env import AlreadySteppingError, NotSteppingError, VecEnv, VecEnvWrapper, \
    CloudpickleWrapper
from mrl.utils.vec_env.dummy_vec_env import DummyVecEnv
from mrl.utils.vec_env.subproc_vec_env import SubprocVecEnv
from mrl.utils.vec_env.vec_frame_stack import VecFrameStack
from mrl.utils.vec_env.vec_normalize import VecNormalize



import numpy as np
import torch
import pickle
import os
import glob
import shutil
from typing import Iterable, Optional, List, Union, Callable
from mrl.utils.misc import short_timestamp
from mrl.utils.misc import AttrDict


class Agent():
  """
  The base agent class. Important: Agents should almost always be generated from a config_dict
  using mrl.util.config_to_agent(config_dict). See configs folder for default configs / examples.
  
  Agent is a flat collection of mrl.Module, which may include:
    - environments (train/eval)
    - replay buffer(s)
    - new task function
    - action function  (exploratory + greedy) 
    - loss function
    - intrinsic curiosity module 
    - value / policy networks and other models (e.g. goal generation)
    - planner (e.g., MCTS)
    - logger
    - anything else you want (image tagger, human interface, etc.)

  Agent has some lifecycle methods (process_experience, optimize, save, load) that call the 
  corresponding lifecycle hooks on modules that declare them.

  Modules have a reference to the Agent so that they can access each other via the Agent. Actually,
  modules use __getattr__ to access the agent directly (via self.*), so they are effectively agent
  methods that are defined in separate files / have their own initialize/save/load functions.

  Modules are registered and saved/restored individually. This lets you swap out / tweak individual
  agent methods without subclassing the agent. Individual saves let you swap out saved modules via
  the filesystem (good for, e.g., BatchRL), avoid pickling problems from non-picklable modules.
  """

  def __init__(
      self,
      module_list: Iterable,  # list of mrl.Modules (possibly nested)
      config: AttrDict):  # hyperparameters and module settings

    self.config = config
    parent_folder = config.parent_folder
    assert parent_folder, "Setting the agent's parent folder is required!"
    self.agent_name = config.get('agent_name') or 'agent_' + short_timestamp()
    self.agent_folder = os.path.join(parent_folder, self.agent_name)
    load_agent = False
    if os.path.exists(self.agent_folder):
      print('Detected existing agent! Loading agent from checkpoint...')
      load_agent = True
    else:
      os.makedirs(self.agent_folder, exist_ok=True)

    self._process_experience_registry = []  # set of modules which define _process_experience
    self._optimize_registry = []            # set of modules which define _optimize
    self.config.env_steps = 0
    self.config.opt_steps = 0

    module_list = flatten_modules(module_list)
    self.module_dict = AttrDict()
    for module in module_list:
      assert module.module_name
      setattr(self, module.module_name, module)
      self.module_dict[module.module_name] = module
    for module in module_list:
      self._register_module(module)

    self.training = True

    if load_agent:
      self.load()
      print('Successfully loaded saved agent!')
    else:
      self.save()

  def train_mode(self):
    """Set agent to train mode; exploration / use dropout / etc. As in Pytorch."""
    self.training = True

  def eval_mode(self):
    """Set agent to eval mode; act deterministically / don't use dropout / etc."""
    self.training = False

  def process_experience(self, experience: AttrDict):
    """Calls the _process_experience function of each relevant module
    (typically, these will include a replay buffer and one or more logging modules)"""
    self.config.env_steps += self.env.num_envs if hasattr(self, 'env') else 1
    for module in self._process_experience_registry:
      module._process_experience(experience)

  def optimize(self):
    """Calls the _optimize function of each relevant module
    (typically, this will be the main algorithm; but may include others)"""
    self.config.opt_steps += 1
    for module in self._optimize_registry:
      module._optimize()

  def _register_module(self, module):
    """
    Provides module with a reference to agent so that modules can interact; e.g., 
    allows agent's policy to reference the value function.

    Then, calls each module's _setup and verify methods to _setup the module and
    verify that agent has all required modules.
    """
    self.module_dict[module.module_name] = module

    module.agent = self
    module.verify_agent_compatibility()
    module._setup()
    module.new_task()
    if hasattr(module, '_process_experience'):
      self._process_experience_registry.append(module)
    if hasattr(module, '_optimize'):
      self._optimize_registry.append(module)

  def set_module(self, module_name, module):
    """
    Sets a module (can be used to switch environments / policies)
    """
    setattr(self, module_name, module)
    self._register_module(module)

  def save(self, subfolder: Optional[str] = None):
    """
    The state of all stateful modules is saved to the agent's folder.
    The agent itself is NOT saved, and should be (1) rebuilt, and (2) restored using self.load().
    Subfolder can be used to save various checkpoints of same agent.
    """
    save_folder = self.agent_folder
    subfolder = subfolder or 'checkpoint'
    save_folder = os.path.join(save_folder, subfolder)

    if not os.path.exists(save_folder):
      os.makedirs(save_folder)

    for module in self.module_dict.values():
      module.save(save_folder)
    
    with open(os.path.join(save_folder, 'config.pickle'), 'wb') as f:
      pickle.dump(self.config, f)

  def load(self, subfolder: Optional[str] = None):
    """
    Restores state of stateful modules from the agent's folder[/subfolder].
    """
    save_folder = self.agent_folder
    subfolder = subfolder or 'checkpoint'
    save_folder = os.path.join(save_folder, subfolder)

    assert os.path.exists(save_folder), "load path does not exist!"
    
    with open(os.path.join(save_folder, 'config.pickle'), 'rb') as f:
      self.config = pickle.load(f)

    for module in self.module_dict.values():
      print("Loading module {}".format(module.module_name))
      module.load(save_folder)

  def save_checkpoint(self, checkpoint_dir):
    """
    Saves agent together with its buffer regardless of save buffer.
    Keeps 2 saves in the in folder in case the job is killed and last
    checkpoint is corrupted.

    NOTE: You should call agent.save to save to the main folder BEFORE calling this.
    """
    if not os.path.exists(checkpoint_dir):
      os.makedirs(checkpoint_dir)

    with open(os.path.join(checkpoint_dir, 'INITIALIZED'), 'w') as f:
      f.write('INITIALIZED')

    subfolder1 = os.path.join(checkpoint_dir, '1')
    subfolder2 = os.path.join(checkpoint_dir, '2')

    os.makedirs(os.path.join(subfolder1, 'checkpoint'), exist_ok=True)
    os.makedirs(os.path.join(subfolder2, 'checkpoint'), exist_ok=True)

    done1 = os.path.join(subfolder1, 'DONE')
    done2 = os.path.join(subfolder2, 'DONE')

    if not os.path.exists(done1):
      savedir = subfolder1
      done_file = done1
    elif not os.path.exists(done2):
      savedir = subfolder2
      done_file = done2
    else:
      modtime1 = os.path.getmtime(done1)
      modtime2 = os.path.getmtime(done2)
      if modtime1 < modtime2:
        savedir = subfolder1
        done_file = done1
      else:
        savedir = subfolder2
        done_file = done2

      os.remove(done_file)

    savedir_checkpoint = os.path.join(savedir, 'checkpoint')
    # First save all modules, including replay buffer
    old_save_replay_buf = self.config.save_replay_buf
    self.config.save_replay_buf = True
    for module in self.module_dict.values():
      module.save(savedir_checkpoint)
    self.config.save_replay_buf = old_save_replay_buf
    
    # Now save the config also
    with open(os.path.join(savedir_checkpoint, 'config.pickle'), 'wb') as f:
      pickle.dump(self.config, f)
    
    # Now copy over the config and results files from the agent_folder
    files_and_folders = glob.glob(os.path.join(self.agent_folder, '*'))
    for file_or_folder in files_and_folders:
      if os.path.isfile(file_or_folder):
        shutil.copy(file_or_folder, savedir)

    # Finally, print the DONE file.
    with open(done_file, 'w') as f:
      f.write('DONE')

  def load_from_checkpoint(self, checkpoint_dir):
    """
    This loads an agent from a checkpoint_dir to which it was saved using the `save_checkpoint` method.
    """
    subfolder1 = os.path.join(checkpoint_dir, '1')
    subfolder2 = os.path.join(checkpoint_dir, '2')
    done1 = os.path.join(subfolder1, 'DONE')
    done2 = os.path.join(subfolder2, 'DONE')

    if not os.path.exists(done1):
      assert os.path.exists(done2)
      savedir = subfolder2
    elif not os.path.exists(done2):
      savedir = subfolder1
    else:
      modtime1 = os.path.getmtime(done1)
      modtime2 = os.path.getmtime(done2)
      if modtime1 > modtime2:
        savedir = subfolder1
      else:
        savedir = subfolder2

    savedir_checkpoint = os.path.join(savedir, 'checkpoint')

    # First load the agent
    with open(os.path.join(savedir_checkpoint, 'config.pickle'), 'rb') as f:
      self.config = pickle.load(f)

    for module in self.module_dict.values():
      print("Loading module {}".format(module.module_name))
      module.load(savedir_checkpoint)

    # Then copy over the config and results file to the agent_folder
    files_and_folders = glob.glob(os.path.join(savedir, '*'))
    for file_or_folder in files_and_folders:
      if os.path.isfile(file_or_folder):
        shutil.copy(file_or_folder, self.agent_folder)


  def torch(self, x):
    if isinstance(x, torch.Tensor): return x
    return torch.FloatTensor(x).to(self.config.device)

  def numpy(self, x):
    return x.cpu().detach().numpy()

class Module(object):
  """
  This is the base class / module for Agent modules. Each module must inherit from it to be used
  in an Agent. 

  So that modules can be saved independently, you should access other modules only through the
  self.agent attribute and not create any new references to other modules.See the Agent.save and 
  Agent.load methods. 
  
  Note that __getattr__ passes through to the agent, so that a call to self.* is the same as a
  call to self.agent.* whenever * is not defined.  
  """

  def __init__(self, module_name: str, 
                     required_agent_modules: Optional[List[str]] = None, 
                     locals: Optional[dict] = None):
    self.module_name = module_name  # Required
    self.config_spec = locals # Optionally used to log arguments to each module (Called by Logger)
    if required_agent_modules is not None:
      self.required_agent_modules = required_agent_modules
    else:
      self.required_agent_modules = []

  def __getattr__(self, name):
    """Attribute access passes through to agent when local attribute does not exist"""
    return getattr(self.agent, name)

  def verify_agent_compatibility(self):
    """Called by agent to verify that module has everything it needs"""
    assert self.module_name is not None
    assert hasattr(self, 'agent')
    for module in self.required_agent_modules:
      assert hasattr(self.agent, module), 'Agent is missing module {}'.format(module)

  def _setup(self):
    """Called after self.agent is set to do any required _setup with other modules"""
    pass

  def new_task(self):
    """Called during _setup, and also by trainig loop if there is a task switch."""
    pass

  def save(self, save_folder: str):
    """Saves module state (note: reference to agent not available). Only some modules 
    have state that is worth saving (e.g., replays or models)"""
    pass

  def load(self, save_folder: str):
    """Restores individual module state"""
    pass

  def _save_props(self, prop_names : List[str], save_folder: str):
    """Convenience method for saving module attributes"""
    prop_dict = {prop: self.__dict__[prop] for prop in prop_names}
    with open(os.path.join(save_folder, "{}_props.pickle".format(self.module_name)), 'wb') as f:
      pickle.dump(prop_dict, f)

  def _load_props(self, prop_names : List[str], save_folder: str):
    """Convenience method for loading module attributes"""
    with open(os.path.join(save_folder, "{}_props.pickle".format(self.module_name)), 'rb') as f:
      prop_dict = pickle.load(f)
    for k, v in prop_dict.items():
      self.__dict__[k] = v

  def __str__(self):
    return self.__class__.__name__


def config_to_agent(config_dict: dict):
  module_list = []
  config = AttrDict()
  for k, v in config_dict.items():
    if is_module_or_or_module_list(v):
      module_list += flatten_modules(v)
    else:
      config[k] = v

  return Agent(module_list, config)


def is_module_or_or_module_list(item):
  if isinstance(item, Module):
    return True
  elif isinstance(item, Iterable):
    if isinstance(item, str):
      return False
    return is_module_or_or_module_list(next(iter(item)))
  else:
    return False


def flatten_modules(module_list: Union[Iterable, Module]):
  res = []
  if isinstance(module_list, Module):
    return [module_list]
  for module in module_list:
    res += flatten_modules(module)
  return res


class FunctionModule(Module):
  """
  Used to wrap functions in an mrl.Module.
  """

  def __init__(self, function: Callable, name: Optional[str] = None):
    super().__init__(name or function.__name__, required_agent_modules=[])
    self.function = function

  def __call__(self, *args, **kwargs):
    return self.function(*args, **kwargs)


from abc import ABC, abstractmethod
import pickle

import cloudpickle

class AlreadySteppingError(Exception):
    """
    Raised when an asynchronous step is running while
    step_async() is called again.
    """

    def __init__(self):
        msg = 'already running an async step'
        Exception.__init__(self, msg)


class NotSteppingError(Exception):
    """
    Raised when an asynchronous step is not running but
    step_wait() is called.
    """

    def __init__(self):
        msg = 'not running an async step'
        Exception.__init__(self, msg)


class VecEnv(ABC):
    """
    An abstract asynchronous, vectorized environment.

    :param num_envs: (int) the number of environments
    :param observation_space: (Gym Space) the observation space
    :param action_space: (Gym Space) the action space
    """

    def __init__(self, num_envs, observation_space, action_space):
        self.num_envs = num_envs
        self.observation_space = observation_space
        self.action_space = action_space

    @abstractmethod
    def reset(self):
        """
        Reset all the environments and return an array of
        observations, or a dict of observation arrays (for goal envs).

        If step_async is still doing work, that work will
        be cancelled and step_wait() should not be called
        until step_async() is invoked again.

        :return: ([int] or [float] or dict) observation
        """
        pass

    @abstractmethod
    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.

        You should not call this if a step_async run is
        already pending.
        """
        pass

    @abstractmethod
    def step_wait(self):
        """
        Wait for the step taken with step_async().

        :return: ([int] or [float] or dict, [float], [bool], dict) observation, reward, done, information
        """
        pass

    @abstractmethod
    def close(self):
        """
        Clean up the environment's resources.
        """
        pass

    def step(self, actions):
        """
        Step the environments with the given action

        :param actions: ([int] or [float]) the action
        :return: ([int] or [float] or dict, [float], [bool], dict) observation, reward, done, information
        """
        self.step_async(actions)
        return self.step_wait()

    def get_images(self):
        """
        Return RGB images from each environment
        """
        raise NotImplementedError

    def render(self, *args, **kwargs):
        """
        Gym environment rendering

        :param mode: (str) the rendering type
        """
        pass

    @property
    def unwrapped(self):
        if isinstance(self, VecEnvWrapper):
            return self.venv.unwrapped
        else:
            return self

    def _get_indices(self, indices):
        """
        Convert a flexibly-typed reference to environment indices to an implied list of indices.
        :param indices: (None,int,Iterable) refers to indices of envs.
        :return: (list) the implied list of indices.
        """
        if indices is None:
            indices = range(self.num_envs)
        elif isinstance(indices, int):
            indices = [indices]
        return indices


class VecEnvWrapper(VecEnv):
    """
    Vectorized environment base class

    :param venv: (VecEnv) the vectorized environment to wrap
    :param observation_space: (Gym Space) the observation space (can be None to load from venv)
    :param action_space: (Gym Space) the action space (can be None to load from venv)
    """

    def __init__(self, venv, observation_space=None, action_space=None):
        self.venv = venv
        VecEnv.__init__(self, num_envs=venv.num_envs, observation_space=observation_space or venv.observation_space,
                        action_space=action_space or venv.action_space)

    def step_async(self, actions):
        self.venv.step_async(actions)

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def step_wait(self):
        pass

    def close(self):
        return self.venv.close()

    def render(self, *args, **kwargs):
        return self.venv.render(*args, **kwargs)

    def get_images(self):
        return self.venv.get_images()

class CloudpickleWrapper(object):
    def __init__(self, var):
        """
        Uses cloudpickle to serialize contents (otherwise multiprocessing tries to use pickle)

        :param var: (Any) the variable you wish to wrap for pickling with cloudpickle
        """
        self.var = var

    def __getstate__(self):
        return cloudpickle.dumps(self.var)

    def __setstate__(self, obs):
        self.var = pickle.loads(obs)



from pathlib import Path

from pydantic import BaseModel


class Paths(BaseModel):

    home: Path = Path(__file__).parents[1]

    configs: Path = home / 'configs'
    configs_rl = configs / 'rl'
    configs_rrt = configs / 'rrt'
    configs_sl = configs / 'sl'
    configs_hyper = configs / 'hyper'

    data: Path = home / 'data'
    data_rl: Path = data / 'rl'
    data_rl_distance: Path = data_rl / 'distance'
    data_sl: Path = data / 'sl'
    data_sl_narrow: Path = data_sl / 'narrow'
    data_sl_narrow_train_csv: Path = data_sl_narrow / 'train.csv'
    data_sl_narrow_val_csv: Path = data_sl_narrow / 'val.csv'

    deps: Path = home / 'deps'
    deps_libraries: Path = deps / 'libraries'
    deps_models: Path = deps / 'models'

    environments: Path = home / 'environments'

    experiments: Path = home / 'experiments'
    experiments_rl: Path = experiments / 'rl'
    experiments_sl: Path = experiments / 'sl'
    experiments_rrt: Path = experiments / 'rrt'

    utilities: Path = home / 'utilities'

    def cfg(self, alg: str, name: str, hyper: bool):
        if hyper:
            return self.configs_hyper / (name + '.yaml')
        else:
            match alg:
                case 'rl':
                    path = self.configs_rl / (name + '.yaml')
                case 'sl':
                    path = self.configs_sl / (name + '.yaml')
                case 'rrt':
                    path = self.configs_rrt / (name + '.yaml')
                case _:
                    raise NotImplementedError
            return path

    def exp(self, alg: str, name: str):
        match alg:
            case 'rl':
                path = self.experiments_rl / name
            case 'sl':
                path = self.experiments_sl / name
            case 'rrt':
                path = self.experiments_rrt / name
            case _:
                raise NotImplementedError
        return path

    def exp_fig(self, alg: str, name: str):
        return self.exp(alg=alg, name=name) / 'figs'

    def exp_tensorboard(self, alg: str, name: str):
        match alg:
            case 'rl':  # In MRL tensorboard file is on the parent folder
                return self.exp(alg=alg, name=name)
            case _:
                return self.exp(alg=alg, name=name) / 'tensorboard'

    def exp_checkpoint(self, alg: str, name: str):
        return self.exp(alg=alg, name=name) / 'checkpoint'

    def exp_stats(self, alg: str, name: str):
        return self.exp(alg=alg, name=name) / 'stats'


class PathsIntellisense:
    """
    Visualize all paths and files.
    """

    def __init__(self, *sub_folders):
        """
        Recursively build a list of paths and files
        :param sub_folders: for recursive calls
        """

        home = Path().cwd().parent
        for folder in sub_folders:
            home = home / folder

        self.path = home

        skip_folders = ['.git', '.idea']
        files_and_dirs = [f for f in list(home.glob('*'))]
        for file_or_dir in files_and_dirs:

            name = str(file_or_dir).rsplit("\\")[-1]
            if name in skip_folders:
                continue

            elif not file_or_dir.is_dir():
                name = name.replace('.', '')
                self.__setattr__(name, file_or_dir)

            else:
                call_sub_folders = [f for f in sub_folders]
                call_sub_folders.append(name)
                self.__setattr__(name, PathsIntellisense(*call_sub_folders))

    def __call__(self, *args, **kwargs):
        """
        Call object to get the path of folders.
        :param args: None.
        :param kwargs: None.
        :return: path to folder.
        """

        return self.path





import pickle

import numpy as np

from mrl.utils.vec_env import VecEnvWrapper

class RunningMeanStd(object):
    def __init__(self, epsilon=1e-4, shape=()):
        """
        calulates the running mean and std of a data stream
        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm

        :param epsilon: (float) helps with arithmetic issues
        :param shape: (tuple) the shape of the data stream's output
        """
        self.mean = np.zeros(shape, 'float64')
        self.var = np.ones(shape, 'float64')
        self.count = epsilon

    def update(self, arr):
        batch_mean = np.mean(arr, axis=0)
        batch_var = np.var(arr, axis=0)
        batch_count = arr.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(self, batch_mean, batch_var, batch_count):
        delta = batch_mean - self.mean
        tot_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / tot_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        m_2 = m_a + m_b + np.square(delta) * self.count * batch_count / (self.count + batch_count)
        new_var = m_2 / (self.count + batch_count)

        new_count = batch_count + self.count

        self.mean = new_mean
        self.var = new_var
        self.count = new_count



class VecNormalize(VecEnvWrapper):
    """
    A moving average, normalizing wrapper for vectorized environment.
    has support for saving/loading moving average,

    :param venv: (VecEnv) the vectorized environment to wrap
    :param training: (bool) Whether to update or not the moving average
    :param norm_obs: (bool) Whether to normalize observation or not (default: True)
    :param norm_reward: (bool) Whether to normalize rewards or not (default: False)
    :param clip_obs: (float) Max absolute value for observation
    :param clip_reward: (float) Max value absolute for discounted reward
    :param gamma: (float) discount factor
    :param epsilon: (float) To avoid division by zero
    """

    def __init__(self, venv, training=True, norm_obs=True, norm_reward=True,
                 clip_obs=10., clip_reward=10., gamma=0.99, epsilon=1e-8):
        VecEnvWrapper.__init__(self, venv)
        self.obs_rms = RunningMeanStd(shape=self.observation_space.shape)
        self.ret_rms = RunningMeanStd(shape=())
        self.clip_obs = clip_obs
        self.clip_reward = clip_reward
        # Returns: discounted rewards
        self.ret = np.zeros(self.num_envs)
        self.gamma = gamma
        self.epsilon = epsilon
        self.training = training
        self.norm_obs = norm_obs
        self.norm_reward = norm_reward
        self.old_obs = np.array([])

    def step_wait(self):
        """
        Apply sequence of actions to sequence of environments
        actions -> (observations, rewards, news)

        where 'news' is a boolean vector indicating whether each element is new.
        """
        obs, rews, news, infos = self.venv.step_wait()
        self.ret = self.ret * self.gamma + rews
        self.old_obs = obs
        obs = self._normalize_observation(obs)
        if self.norm_reward:
            if self.training:
                self.ret_rms.update(self.ret)
            rews = np.clip(rews / np.sqrt(self.ret_rms.var + self.epsilon), -self.clip_reward, self.clip_reward)
        self.ret[news] = 0
        return obs, rews, news, infos

    def _normalize_observation(self, obs):
        """
        :param obs: (numpy tensor)
        """
        if self.norm_obs:
            if self.training:
                self.obs_rms.update(obs)
            obs = np.clip((obs - self.obs_rms.mean) / np.sqrt(self.obs_rms.var + self.epsilon), -self.clip_obs,
                          self.clip_obs)
            return obs
        else:
            return obs

    def get_original_obs(self):
        """
        returns the unnormalized observation

        :return: (numpy float)
        """
        return self.old_obs

    def reset(self):
        """
        Reset all environments
        """
        obs = self.venv.reset()
        if len(np.array(obs).shape) == 1:  # for when num_cpu is 1
            self.old_obs = [obs]
        else:
            self.old_obs = obs
        self.ret = np.zeros(self.num_envs)
        return self._normalize_observation(obs)

    def save_running_average(self, path):
        """
        :param path: (str) path to log dir
        """
        for rms, name in zip([self.obs_rms, self.ret_rms], ['obs_rms', 'ret_rms']):
            with open("{}/{}.pkl".format(path, name), 'wb') as file_handler:
                pickle.dump(rms, file_handler)

    def load_running_average(self, path):
        """
        :param path: (str) path to log dir
        """
        for name in ['obs_rms', 'ret_rms']:
            with open("{}/{}.pkl".format(path, name), 'rb') as file_handler:
                setattr(self, name, pickle.load(file_handler))



import mrl
from mrl.utils.misc import AttrDict
import numpy as np

class EpisodicEval(mrl.Module):
  def __init__(self):
    super().__init__('eval', required_agent_modules = ['eval_env', 'policy'], locals=locals())
  
  def __call__(self, num_episodes : int, *unused_args):
    """
    Runs num_steps steps in the environment and returns results.
    Results tracking is done here instead of in process_experience, since 
    experiences aren't "real" experiences; e.g. agent cannot learn from them.  
    """
    self.eval_mode()
    env = self.eval_env
    num_envs = env.num_envs
    
    episode_rewards, episode_steps = [], []
    discounted_episode_rewards = []
    is_successes = []
    record_success = False

    while len(episode_rewards) < num_episodes:
      state = env.reset()

      dones = np.zeros((num_envs,))
      steps = np.zeros((num_envs,))
      is_success = np.zeros((num_envs,))
      ep_rewards = [[] for _ in range(num_envs)]

      while not np.all(dones):
        action = self.policy(state)
        state, reward, dones_, infos = env.step(action)

        for i, (rew, done, info) in enumerate(zip(reward, dones_, infos)):
          if dones[i]:
            continue
          ep_rewards[i].append(rew)
          steps[i] += 1
          if done:
            dones[i] = 1. 
            if 'is_success' in info:
              record_success = True
              is_success[i] = info['is_success']

      for ep_reward, step, is_succ in zip(ep_rewards, steps, is_success):
        if record_success:
          is_successes.append(is_succ)
        episode_rewards.append(sum(ep_reward))
        discounted_episode_rewards.append(discounted_sum(ep_reward, self.config.gamma))
        episode_steps.append(step)
    
    if hasattr(self, 'logger'):
      if len(is_successes):
        self.logger.add_scalar('Test/Success', np.mean(is_successes))
      self.logger.add_scalar('Test/Episode_rewards', np.mean(episode_rewards))
      self.logger.add_scalar('Test/Discounted_episode_rewards', np.mean(discounted_episode_rewards))
      self.logger.add_scalar('Test/Episode_steps', np.mean(episode_steps))

    return AttrDict({
      'rewards': episode_rewards,
      'steps': episode_steps
    })

def discounted_sum(lst, discount):
  sum = 0
  gamma = 1
  for i in lst:
    sum += gamma*i
    gamma *= discount
  return sum


import torch
import pytest
import numpy as np
from torch.utils.data import TensorDataset, DataLoader

from torchensemble.utils.dataloder import FixedDataLoader


# Data
X = torch.Tensor(np.array(([0.1, 0.1], [0.2, 0.2], [0.3, 0.3], [0.4, 0.4])))
y = torch.LongTensor(np.array(([0, 0, 1, 1])))

data = TensorDataset(X, y)
dataloder = DataLoader(data, batch_size=2, shuffle=False)


def test_fixed_dataloder():
    fixed_dataloader = FixedDataLoader(dataloder)
    for _, (fixed_elem, elem) in enumerate(zip(fixed_dataloader, dataloder)):
        # Check same elements
        for elem_1, elem_2 in zip(fixed_elem, elem):
            assert torch.equal(elem_1, elem_2)

    # Check dataloder length
    assert len(fixed_dataloader) == 2


def test_fixed_dataloader_invalid_type():
    with pytest.raises(ValueError) as excinfo:
        FixedDataLoader((X, y))
    assert "input used to instantiate FixedDataLoader" in str(excinfo.value)



from coral_pytorch.losses import corn_loss

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.hints import *


def get_ackermann_v_rf_lr_phi_lr(v_ref: float, phi_ref: float, cfg: MasterConfig) -> Vector6:
    """
    Ackermann formula to get velocity for each wheel and steering angles.
    :param v_ref: desired linear velocity
    :param phi_ref: desired steering angle
    :param cfg: master config that contains car parameters
    :return: [vrl vrr vfl vfr phil phir] (f: forward, r: rear/right, l: left)
    """

    # Clip maximum steering angle
    if abs(phi_ref) > cfg.env.car.phi_max:
        phi_ref = phi_ref / (abs(phi_ref)) * cfg.env.car.phi_max

    # V to rad/s
    v_ref = np.clip(v_ref, -cfg.env.car.v_max, +cfg.env.car.v_max)
    v_ref = v_ref / cfg.env.car.wheel_radius
    if phi_ref == 0:
        r = 99999999
    else:
        r = cfg.env.car.axis_dist / (np.tan(phi_ref))

    # Right and left wheel
    phir = np.arctan(cfg.env.car.axis_dist / (r + cfg.env.car.wheel_dist / 2))
    phil = np.arctan(cfg.env.car.axis_dist / (r - cfg.env.car.wheel_dist / 2))

    # Target velocity for wheels
    vrr = v_ref * (1 + cfg.env.car.wheel_dist / (2 * r))
    vrl = v_ref * (1 - cfg.env.car.wheel_dist / (2 * r))
    vfr = vrr / (np.cos(phir))
    vfl = vrl / (np.cos(phil))

    return vrl, vrr, vfl, vfr, phil, phir


def quaternion_to_theta(quaternion: Vector4) -> float:
    """
    Convert quaternion to theta (orientation in the xy-plane)
    :param quaternion: orientation from bullet
    :return: theta angle
    """

    w = quaternion[3]
    theta = 2 * np.arccos(w)
    theta = theta * quaternion[2] / np.abs(quaternion[2]) if quaternion[2] != 0 else theta

    return theta


def path_length(xs, ys):
    """
    Get path length using arc length formula.
    :param xs: list of xs
    :param ys: list of ys
    :return: path length
    """

    d = 0
    for i, _ in enumerate(zip(xs[:-1], ys[:-1])):
        d += np.sqrt((xs[i]-xs[i+1])**2 + (ys[i]-ys[i+1])**2)

    return d


def pose3_to_state4(pose: Vector3) -> Vector4:
    """
    Unpack angle into sine and cosine, maintaining x and y coordinates.
    :param pose: [x y theta]
    :return: [x y sin_theta cos_theta]
    """

    return np.array([pose[0], pose[1], np.sin(pose[2]), np.cos(pose[2])])


def pose3_to_state4_sl(pose: Vector3 | np.ndarray) -> Vector4 | np.ndarray:
    """
    State in SL is in the form [x y cos sin]
    :param pose: pose as [x y theta]
    :return: state in SL form
    """

    # ndarray  case

    if isinstance(pose, np.ndarray):

        if len(pose.shape) == 1:
            cos = np.cos(pose[2])
            sin = np.sin(pose[2])
            state_sl = np.array([pose[0], pose[1], cos, sin])

        elif len(pose.shape) == 2:
            cos = np.cos(pose[:, 2])
            sin = np.sin(pose[:, 2])
            state_sl = np.zeros(shape=(pose.shape[0], 4))
            state_sl[:, 0] = pose[:, 0]
            state_sl[:, 1] = pose[:, 1]
            state_sl[:, 2] = cos
            state_sl[:, 3] = sin

        else:
            raise NotImplementedError
    else:
        state_sl = pose3_to_state4_sl(np.array(pose))

    return state_sl


def scale_to_range(x: float, src_range: Vector, dst_range: Vector):
    """
    Scale a number from a source range to a destination range.
    :param x: input number
    :param src_range: source range
    :param dst_range: destination range
    :return: scaled vector
    """

    return (x - src_range[0]) * (dst_range[1] - dst_range[0]) / (src_range[1] - src_range[0]) + dst_range[0]


def state4_to_pose3(state: Vector4) -> Vector3:
    """
    Convert [x y sin_theta cos_theta] to [x y theta].
    :param state: vector with unpacked angle
    :return: pose
    """

    return np.array([state[0], state[1], np.arctan2(state[2], state[3])])


def state6_to_pose3(state: Vector6) -> Vector3:
    """
    Convert [x y sin_theta cos_theta v phi] to [x y theta]
    :param state: vector with 6 components, including velocity and steering angle
    :return: corresponding position and orientation
    """

    return np.array([state[0], state[1], np.arctan2(state[2], state[3])])


def state4_sl_to_state4_rl(state_sl: Vector4 | np.ndarray) -> Vector4 | np.ndarray:
    """
    State in SL is in the form [x y cos sin], and in RL is [x y sin cos]
    :param state_sl: state on SL dataset
    :return: state in RL form
    """

    # Single state case
    if len(state_sl.shape) == 1:
        cos = state_sl[2]
        sin = state_sl[3]
        state_rl = np.array([state_sl[0], state_sl[1], sin, cos])

    elif len(state_sl.shape) == 2:
        cos = state_sl[:, 2]
        sin = state_sl[:, 3]
        state_rl = np.copy(state_sl)
        state_rl[:, 2] = sin
        state_rl[:, 3] = cos

    else:
        raise NotImplementedError

    return state_rl


def state4_sl_to_pose3(state_sl: Vector4 | np.ndarray) -> Vector3 | np.ndarray:
    """
    State in SL is in the form [x y cos sin], so we convert to [x y theta]
    :param state_sl: state on SL dataset
    :return: pose [x y theta]
    """

    # Single state case
    if len(state_sl.shape) == 1:
        cos = state_sl[2]
        sin = state_sl[3]
        state_rl = np.array([state_sl[0], state_sl[1], np.arctan2(sin, cos)])

    elif len(state_sl.shape) == 2:
        cos = state_sl[:, 2]
        sin = state_sl[:, 3]
        state_rl = np.zeros(shape=(state_sl.shape[0], 3))
        state_rl[:, 0] = state_sl[:, 0]
        state_rl[:, 1] = state_sl[:, 1]
        state_rl[:, 2] = np.arctan2(sin, cos)

    else:
        raise NotImplementedError

    return state_rl


def transform_to_origin(obs=None, node_from=None, node_to=None):
    """
    Transform target pose to relative pose wrt car.
    :param obs: env observation
    :param node_from: node of car origin
    :param node_to: node of target pose
    :return: new target pose
    """

    # Handle args and get poses
    if obs is None:
        assert node_from is not None and node_to is not None
        car_pose = node_from.pose
        target_pose = node_to.pose
    else:
        assert node_from is None and node_to is None
        car_pose = state4_to_pose3(obs['achieved_goal'])
        target_pose = state4_to_pose3(obs['desired_goal'])
    
    # Constants
    x = target_pose[0]
    y = target_pose[1]
    ct = np.cos(-car_pose[2])
    st = np.sin(-car_pose[2])

    # Translate target
    x = x - car_pose[0]
    y = y - car_pose[1]

    # 2D rotation
    x_new = x*ct - y*st
    y_new = y*ct + x*st
    theta_new = target_pose[2] - car_pose[2]

    return np.array([x_new, y_new, theta_new])





import mrl
import numpy as np

# temporary fix for dumb tensorflow / tensorboard error. https://github.com/pytorch/pytorch/issues/30966
try:
  import tensorflow as tf
  import tensorboard as tb
  tf.io.gfile = tb.compat.tensorflow_stub.io.gfile
except:
  import tensorboard as tb

from torch.utils.tensorboard import SummaryWriter
from tabulate import tabulate
from collections import defaultdict
import json
import os
import time
import csv

class Logger(mrl.Module):
  """
  Logger that processes vectorized experiences and 
  records results to the console. 
  """
  def __init__(self, average_every=100):
    super().__init__('logger', required_agent_modules=['env'], locals=locals())
    self.average_every = average_every
    self.writer = None

  def _setup(self):
    # rewards and steps are always tracked
    self.rewards_per_env = np.zeros((self.env.num_envs, ))
    self.steps_per_env = np.zeros((self.env.num_envs, ))
    self.episode_rewards = []
    self.episode_steps = []
    self.steps = 0
    self.episodes = 0
    self.tabular = defaultdict(list)
    self.last_log_step = defaultdict(int)
    self.log_every_n_steps = self.config.log_every
    self.save_config()

  def lazy_init_writer(self):
    if self.writer is None:
      self.writer = SummaryWriter(self.agent_folder)

  def update_csv(self, tag, value, step):
    fields = ['wall_time', 'step', tag]
    path = os.path.join(self.agent_folder, self.agent_name + '__' + tag.replace('/', '__') + '.csv')
    if not os.path.exists(path):
      with open(path, 'w') as f:
        writer = csv.writer(f, delimiter=',')
        writer.writerow(fields)
    with open(path, 'a') as f:
      writer = csv.writer(f)
      writer.writerow([time.time(), step, value])

  def add_scalar(self, tag, value, log_every=1000, step=None):
    """Adds scalar to tensorboard"""
    self.lazy_init_writer()
    if step is None:
      step = self.config.env_steps
    if step - self.last_log_step[tag] >= log_every:
      self.last_log_step[tag] = step
      self.writer.add_scalar(tag, value, step)
      self.update_csv(tag, value, step)

  def add_histogram(self, tag, values, log_every=1000, step=None, **kwargs):
    """Adds histogram to tensorboard"""
    self.lazy_init_writer()
    if isinstance(values, list):
      values = np.array(values, dtype=np.float32)
    elif isinstance(values, np.ndarray):
      values = values.astype(np.float32)
    if step is None:
      step = self.config.env_steps
    if step - self.last_log_step[tag] >= log_every:
      self.last_log_step[tag] = step
      self.writer.add_histogram(tag, values, step, **kwargs)

  def add_embedding(self, tag, values, log_every=1000, step=None, **kwargs):
    """Adds embedding data to tensorboard"""
    self.lazy_init_writer()
    if isinstance(values, list):
      values = np.array(values, dtype=np.float32)
    elif isinstance(values, np.ndarray):
      values = values.astype(np.float32)
    assert len(values.shape) == 2
    if step is None:
      step = self.config.env_steps
    if step - self.last_log_step[tag] >= log_every:
      self.last_log_step[tag] = step
      self.writer.add_embedding(mat=values, tag=tag, global_step=step, **kwargs)

  def add_tabular(self, tag, value):
    """Adds scalar to console logger"""
    self.tabular[tag].append(value)

  def log_color(self, tag, value='', color='cyan'):
    print(colorize(tag, color=color, bold=True), value)

  def save_config(self):
    config_json = convert_json({**self.config, **record_attrs(self.module_dict.values())})
    config_json['agent_name'] = self.agent_name
    output = json.dumps(config_json, separators=(',', ':\t'), indent=4, sort_keys=True)
    print(colorize('\nAgent folder:', color='magenta', bold=True))
    print(self.agent_folder)
    print(colorize('\nSaving config:', color='cyan', bold=True))
    print(output)
    with open(os.path.join(self.agent_folder, "config.json"), 'w') as out:
      out.write(output)

  def flush_console(self):
    table = [('Environment steps', self.steps), ('Total episodes', self.episodes),
             ('Avg rewards (last {})'.format(self.average_every), np.mean(self.episode_rewards[-self.average_every:])),
             ('Avg episode len (last {})'.format(self.average_every), np.mean(self.episode_steps[-self.average_every:]))
             ]
    for k, v in self.tabular.items():
      table.append(('Avg ' + k + ' (last {})'.format(self.average_every), np.mean(v[-self.average_every:])))
    table = tabulate(table, headers=['Tag', 'Value'], tablefmt="psql", floatfmt="8.1f")

    print(table)

  def _process_experience(self, experience):
    rewards, dones = experience.reward, experience.trajectory_over

    self.rewards_per_env += rewards
    self.steps_per_env += 1

    if np.any(dones):
      self.episode_rewards += list(self.rewards_per_env[dones])
      self.episode_steps += list(self.steps_per_env[dones])
      self.rewards_per_env[dones] = 0
      self.steps_per_env[dones] = 0
      self.episodes += np.sum(dones)

    self.steps += self.env.num_envs
    if self.steps % self.log_every_n_steps < self.env.num_envs:
      self.flush_console()
      self.add_scalar('Train/Episode_rewards', np.mean(self.episode_rewards[-30:]))
      self.add_scalar('Train/Episode_steps', np.mean(self.episode_steps[-30:]))

  def save(self, save_folder):
    self._save_props([
        'episode_rewards', 'episode_steps',
        'steps', 'episodes', 'tabular', 'last_log_step'
    ], save_folder)

  def load(self, save_folder):
    self._load_props([
        'episode_rewards', 'episode_steps',
        'steps', 'episodes', 'tabular', 'last_log_step'
    ], save_folder)


color2num = dict(gray=30, red=31, green=32, yellow=33, blue=34, magenta=35, cyan=36, white=37, crimson=38)


def colorize(string, color, bold=False, highlight=False):
  """
    Colorize a string.
    This function was originally written by John Schulman.
    """
  attr = []
  num = color2num[color]
  if highlight: num += 10
  attr.append(str(num))
  if bold: attr.append('1')
  return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), string)


def convert_json(obj, dict_to_str=False):
  """ Convert obj to a version which can be serialized with JSON. """
  if is_json_serializable(obj):
    return obj
  else:
    if isinstance(obj, dict) and not dict_to_str:
      return {convert_json(k): convert_json(v) for k, v in obj.items()}

    elif isinstance(obj, tuple):
      return tuple([convert_json(x) for x in obj])

    elif isinstance(obj, list):
      return [convert_json(x) for x in obj]

    elif hasattr(obj, '__name__') and not ('lambda' in obj.__name__):
      return convert_json(obj.__name__)

    elif hasattr(obj, '__dict__') and obj.__dict__:
      obj_dict = {convert_json(k): convert_json(v) for k, v in obj.__dict__.items()}
      return {str(obj): obj_dict}

    return str(obj)


def is_json_serializable(v):
  try:
    json.dumps(v)
    return True
  except:
    return False


def record_attrs(module_list):
  res = {}
  for module in module_list:
    res['module_' + module.module_name] = convert_json(strip_config_spec(module.config_spec), dict_to_str=True)
  return res


def strip_config_spec(config_spec):
  if '__class__' in config_spec:
    del config_spec['__class__']
  return config_spec


import mrl
from mrl.utils.misc import AttrDict
import numpy as np
from copy import deepcopy
import time

class StandardTrain(mrl.Module):
  def __init__(self):
    super().__init__('train', required_agent_modules = ['env', 'policy', 'optimize'], locals=locals())

  def _setup(self):
    assert hasattr(self.config, 'optimize_every')
    self.optimize_every = self.config.optimize_every
    self.env_steps = 0
    self.reset_idxs = []

  def __call__(self, num_steps : int, render=False, dont_optimize=False, dont_train=False):
    """
    Runs num_steps steps in the environment, saves collected experiences,
    and trains at every step
    """
    if not dont_train:
      self.agent.train_mode()
    env = self.env
    state = env.state 

    for _ in range(num_steps // env.num_envs):
      action = self.policy(state)
      next_state, reward, done, info = env.step(action)

      if self.reset_idxs:
        env.reset(self.reset_idxs)
        for i in self.reset_idxs:
          done[i] = True
          if not 'done_observation' in info[i]:
            if isinstance(next_state, np.ndarray):
              info[i].done_observation = next_state[i]
            else:
              for key in next_state:
                info[i].done_observation = {k: next_state[k][i] for k in next_state}
        next_state = env.state
        self.reset_idxs = []

      state, experience = debug_vectorized_experience(state, action, next_state, reward, done, info)
      self.process_experience(experience)

      if render:
        time.sleep(0.02)
        env.render()
      
      for _ in range(env.num_envs):
        self.env_steps += 1
        if self.env_steps % self.optimize_every == 0 and not dont_optimize:
          self.optimize()
    
    # If using MEP prioritized replay, fit the density model
    if self.config.prioritized_mode == 'mep':
      self.prioritized_replay.fit_density_model()
      self.prioritized_replay.update_priority()
  
  def reset_next(self, idxs):
    """Resets specified envs on next step"""
    self.reset_idxs = idxs

  def save(self, save_folder):
    self._save_props(['env_steps'], save_folder)

  def load(self, save_folder):
    self._load_props(['env_steps'], save_folder)

def debug_vectorized_experience(state, action, next_state, reward, done, info):
  """Gym returns an ambiguous "done" signal. VecEnv doesn't 
  let you fix it until now. See ReturnAndObsWrapper in env.py for where
  these info attributes are coming from."""
  experience = AttrDict(
    state = state,
    action = action,
    reward = reward,
    info = info
  )
  next_copy = deepcopy(next_state) # deepcopy handles dict states

  for idx in np.argwhere(done):
    i = idx[0]
    if isinstance(next_copy, np.ndarray):
      next_copy[i] = info[i].done_observation
    else:
      assert isinstance(next_copy, dict)
      for key in next_copy:
        next_copy[key][i] = info[i].done_observation[key]
  
  experience.next_state = next_copy
  experience.trajectory_over = done
  experience.done = np.array([info[i].terminal_state for i in range(len(done))], dtype=np.float32)
  experience.reset_state = next_state
  
  return next_state, experience


import pytest
import torchensemble
import torch.nn as nn


optimizer_list = [
    "Adadelta",
    "Adagrad",
    "Adam",
    "AdamW",
    "Adamax",
    "ASGD",
    "RMSprop",
    "Rprop",
    "SGD",
]


# Base estimator
class MLP(nn.Module):
    def __init__(self):
        super(MLP, self).__init__()
        self.linear1 = nn.Linear(2, 2)
        self.linear2 = nn.Linear(2, 2)

    def forward(self, X):
        X = X.view(X.size()[0], -1)
        output = self.linear1(X)
        output = self.linear2(output)
        return output


@pytest.mark.parametrize("optimizer_name", optimizer_list)
def test_set_optimizer_normal(optimizer_name):
    model = MLP()
    torchensemble.utils.set_module.set_optimizer(
        model, optimizer_name, lr=1e-3
    )


def test_set_optimizer_Unknown():
    model = MLP()

    with pytest.raises(NotImplementedError) as excinfo:
        torchensemble.utils.set_module.set_optimizer(model, "Unknown")
    assert "Unrecognized optimizer" in str(excinfo.value)


def test_update_lr():
    cur_lr = 1e-4
    model = MLP()
    optimizer = torchensemble.utils.set_module.set_optimizer(
        model, "Adam", lr=1e-3
    )

    optimizer = torchensemble.utils.set_module.update_lr(optimizer, cur_lr)

    for group in optimizer.param_groups:
        assert group["lr"] == cur_lr


def test_update_lr_invalid():
    cur_lr = 0
    model = MLP()
    optimizer = torchensemble.utils.set_module.set_optimizer(
        model, "Adam", lr=1e-3
    )

    err_msg = (
        "The learning rate should be strictly positive, but got" " {} instead."
    ).format(cur_lr)
    with pytest.raises(ValueError, match=err_msg):
        torchensemble.utils.set_module.update_lr(optimizer, cur_lr)



from typing import TypeAlias

import numpy as np


Vector: TypeAlias = list[float, ...] | tuple[float, ...] | np.ndarray
Vector2: TypeAlias = list[float, float] | tuple[float, float] | np.ndarray
Vector3: TypeAlias = list[float, float, float] | tuple[float, float, float] | np.ndarray
Vector4: TypeAlias = list[float, float, float, float] | tuple[float, float, float, float] | np.ndarray
Vector5: TypeAlias = list[float, float, float, float, float] | tuple[float, float, float, float, float] | np.ndarray
Vector6: TypeAlias = list[float, float, float, float, float, float] | tuple[float, float, float, float, float, float] \
                     | np.ndarray



from glob import glob
from pathlib import Path
from typing import Any

import numpy as np
from joblib import dump, load
from pydantic import BaseModel
from scipy.spatial.distance import euclidean

from rrt_ml.utilities.configs import *
from rrt_ml.utilities.hints import *
from rrt_ml.utilities.misc import *


class BaseStats(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    z: None = None

    def save_to_file(self, path: Path | str):
        """
        Save pickle to file.
        """

        # Convert to string if needed
        if isinstance(path, Path):
            path = str(path)

        # Get dict and save to file
        d = self.dict()
        with open(path, 'wb') as f:
            dump(d, f)

        # Split possibly large pickle file
        self._split(path)

        # Delete possibly large pickle file
        Path(path).unlink(missing_ok=True)

    @classmethod
    def load_from_file(cls, path: Path):
        """
        Load from file.
        """

        # Join split files
        cls()._join(str(path))

        # Instantiate class from file dict
        with open(str(path), 'rb') as f:
            d = load(f)
        new = cls(**d)

        # Remove stats file created to load file
        path.unlink(missing_ok=True)

        return new

    @staticmethod
    def _split(path: str):
        """
        Split large file into smaller parts.
        """

        # Open the source file in binary mode
        input_file = open(path, 'rb')

        # Start file counter
        counter = 0

        # Loop read chunks and write to file
        while True:

            # Read a portion of the input file
            chunk = input_file.read(96000000)

            # End the loop if we have hit EOF
            if not chunk:
                break

            # Increment counter
            counter += 1

            # Create a new file name
            filename = path + str(counter).zfill(2)

            # Create a destination file
            dest_file = open(filename, 'wb')

            # Write to this portion of the destination file and close the file
            dest_file.write(chunk)
            dest_file.close()

        # Close the file being read
        input_file.close()

    @staticmethod
    def _join(path: str):
        """
        Join split pickled files.
        """

        # Create a new destination file
        output_file = open(path, 'wb')

        # Get a list of the file parts
        parts = glob(path + '[0-9][0-9]')

        # Go through each portion one by one
        for file in parts:

            # Assemble the full path to the file
            path = file

            # Open the part
            input_file = open(path, 'rb')

            # Read and merge
            while True:

                # Read all bytes of the part
                b = input_file.read(96000000)

                # Break out of loop if we are at end of file
                if not b:
                    break

                # Write the bytes to the output file
                output_file.write(b)

            # Close the input file
            input_file.close()

        # Close the output file
        output_file.close()


class EnvStats(BaseStats):

    class _bullet(BaseStats):
        car_poses: np.ndarray | None = None  # ep_n x timestep x pose (3)
        car_steers: np.ndarray | None = None  # ep_n x timestep x position (2)
        car_velocities: np.ndarray | None = None  # ep_n x timestep x velocity (1)
        dones: np.ndarray | None = None  # ep_n x timestep x done (bool, dict[str, bool])
        rewards: np.ndarray | None = None  # ep_n x timestep x reward (1)
        target_poses: np.ndarray | None = None  # ep_n x timestep x pose (3)
        time: np.ndarray | None = None  # ep_n x timestep x time (1)
        wall_time: np.ndarray | None = None  # ep_n x timestep x wall_time (1)

    class _mdp(BaseStats):
        achieved_goals: np.ndarray | None = None  # ep_n x timestep x goal (4)
        actions: np.ndarray | None = None  # ep_n x timestep x action (2)
        desired_goals: np.ndarray | None = None  # ep_n x timestep x goal (4)
        dones: np.ndarray | None = None  # ep_n x timestep x done (bool, dict[str, bool])
        next_achieved_goals: np.ndarray | None = None  # ep_n x timestep x goal (4)
        next_desired_goals: np.ndarray | None = None  # ep_n x timestep x goal (4)
        next_states: np.ndarray | None = None  # ep_n x timestep x state(6)
        rewards: np.ndarray | None = None  # ep_n x timestep x reward (1)
        states: np.ndarray | None = None  # ep_n x timestep x state (6)

    bullet: _bullet = _bullet()
    mdp: _mdp = _mdp()

    @classmethod
    def new(cls, cfg: MasterConfig):
        # Constants from config
        self = cls()
        n_episodes = cfg.rl.val.n_episodes
        n_mdp_timesteps = cfg.env.general.max_timestep
        n_bullet_timesteps = n_mdp_timesteps * (int(1 / (cfg.env.car.f_val * (1 / 240))))

        # MDP attributes
        self.mdp.states = np.zeros((n_episodes, n_mdp_timesteps, 6))
        self.mdp.achieved_goals = np.zeros((n_episodes, n_mdp_timesteps, 4))
        self.mdp.desired_goals = np.zeros((n_episodes, n_mdp_timesteps, 4))
        self.mdp.next_states = np.zeros((n_episodes, n_mdp_timesteps, 6))
        self.mdp.next_achieved_goals = np.zeros((n_episodes, n_mdp_timesteps, 4))
        self.mdp.next_desired_goals = np.zeros((n_episodes, n_mdp_timesteps, 4))
        self.mdp.actions = np.zeros((n_episodes, n_mdp_timesteps, 2))
        self.mdp.rewards = np.zeros((n_episodes, n_mdp_timesteps, 1))
        self.mdp.dones = np.empty((n_episodes, n_mdp_timesteps, 2), dtype=object)

        # Bullet attributes
        self.bullet.car_poses = np.zeros((n_episodes, n_bullet_timesteps, 3))
        self.bullet.car_steers = np.zeros((n_episodes, n_bullet_timesteps, 2))
        self.bullet.car_velocities = np.zeros((n_episodes, n_bullet_timesteps, 1))
        self.bullet.dones = np.empty((n_episodes, n_bullet_timesteps, 2), dtype=object)
        self.bullet.rewards = np.zeros((n_episodes, n_bullet_timesteps, 1))
        self.bullet.target_poses = np.zeros((n_episodes, n_bullet_timesteps, 3))
        self.bullet.time = np.zeros((n_episodes, n_bullet_timesteps, 1))
        self.bullet.wall_time = np.zeros((n_episodes, n_bullet_timesteps, 1))

        return self

    def get_distance_traveled(self, ep_num: int) -> float:
        """
        Calculate total distance traveled in an episode.
        :param ep_num: episode number
        :return: distance in meters
        """

        # Get xs and ys
        poses = self.bullet.car_poses[ep_num, ...]
        xs = poses[:, 0].flatten()
        ys = poses[:, 1].flatten()

        # Remove zeros
        xs = remove_trailing_zeros(xs)
        ys = remove_trailing_zeros(ys)

        # Sum differentials
        d = 0
        for i, (x, y) in enumerate(zip(xs[:-1], ys[:-1])):
            d += euclidean([x, y], [xs[i + 1], ys[i + 1]])

        return d

    def get_distance_traveled(self, ep_num: int) -> float:
        """
        Calculate total distance traveled in an episode.
        :param ep_num: episode number
        :return: distance in meters
        """

        # Get xs and ys
        poses = self.bullet.car_poses[ep_num, ...]
        xs = poses[:, 0].flatten()
        ys = poses[:, 1].flatten()

        # Remove zeros
        xs = remove_trailing_zeros(xs)
        ys = remove_trailing_zeros(ys)

        # Sum differentials
        d = 0
        for i, (x, y) in enumerate(zip(xs[:-1], ys[:-1])):
            d += euclidean([x, y], [xs[i + 1], ys[i + 1]])

        return d

    def get_time_to_reach(self, ep_num: int):
        """
        Get total time to reach goal in an episode.
        :param ep_num: episode number
        :return: total simulation time to reach
        """

        # Get list of times
        l_times = self.bullet.time[ep_num, :, 0]

        # Get last non-zero element of array (remove zeros and flat)
        return l_times.ravel()[np.flatnonzero(l_times)][-1]


class RLStats(BaseStats):
    det_val_rewards: list[float, ...] | None = None
    rand_val_rewards: list[float, ...] | None = None
    env_stats: list[EnvStats, ...] | None = None
    total_val_rewards: list[float, ...] | None = None
    train_timestep_nums: list[int, ...] | None = None

    @classmethod
    def new(cls, cfg):
        stats = cls()
        stats.det_val_rewards = []
        stats.rand_val_rewards = []
        stats.env_stats = []
        stats.total_val_rewards = []
        stats.train_timestep_nums = []
        return stats


class SLStats(BaseStats):
    arr_epoch_idx_state_dim: None | np.ndarray = None  # epoch x idx x sample_n x state
    train_loss: Vector | None = None
    train_kl_loss: Vector | None = None
    train_recon_loss: Vector | None = None
    val_loss: Vector | None = None
    val_kl_loss: Vector | None = None
    val_recon_loss: Vector | None = None

    @classmethod
    def new(cls, cfg: MasterConfig):
        stats = cls()
        n_epochs = cfg.sl.train.n_epochs
        num_val_maps = cfg.sl.val.n_maps
        num_gen_states = cfg.sl.val.n_states
        state_dim = cfg.sl.dim.state
        stats.arr_epoch_idx_state_dim = np.zeros(shape=(n_epochs, num_val_maps, num_gen_states, state_dim))
        stats.train_loss = []
        stats.train_kl_loss = []
        stats.train_recon_loss = []
        stats.val_loss = []
        stats.val_kl_loss = []
        stats.val_recon_loss = []
        return stats


class RRTStats(BaseStats):
    l_nodes: list[list[Any, ...], ...] | None = None
    l_wall_times: list[float, ...] | None = None

    @classmethod
    def new(cls, cfg):
        stats = cls()
        stats.l_nodes = []
        stats.l_wall_times = []
        return stats



from rrt_ml.algorithms.rrt import *
from rrt_ml.utilities.configs import *


def run_rrt(cfg: MasterConfig):

    # Initialize RRT
    rrt = RRT(cfg)

    # Train
    if cfg.general.is_train:
        rrt.train()

    # Test
    else:
        rrt.test()



"""
Just using Shantong's code for now, but really this should be implemented in 
Pytorch and done on GPU when it is available (see old code for a TF version)
"""

#######################################################################
# Copyright (C) 2017 Shangtong Zhang(zhangshangtong.cpp@gmail.com)    #
# Permission given to modify the code as long as you keep this        #
# declaration at the top                                              #
#######################################################################

import numpy as np
from mrl.utils.schedule import *


class RandomProcess(object):
  def reset_states(self):
    pass


class GaussianProcess(RandomProcess):
  def __init__(self, size, std=ConstantSchedule(0.2)):
    self.size = size
    self.std = std

  def sample(self):
    return np.random.randn(*self.size) * self.std()


class OrnsteinUhlenbeckProcess(RandomProcess):
  def __init__(self, size, std=ConstantSchedule(0.2), theta=.15, dt=1e-2, x0=None, reset_every=50):
    self.theta = theta
    self.mu = 0
    self.std = std
    self.dt = dt
    self.x0 = x0
    self.size = size
    self.reset_states()
    self.steps = 0
    self.reset_every = reset_every

  def sample(self):
    self.steps += 1
    if self.steps % self.reset_every == 0:
      self.reset_states()
    x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + self.std() * np.sqrt(
        self.dt) * np.random.randn(*self.size)
    self.x_prev = x
    return x

  def reset_states(self):
    self.x_prev = self.x0 if self.x0 is not None else np.zeros(self.size)



"""
Some of this is direct or modified from:
https://github.com/ShangtongZhang/DeepRL/tree/master/deep_rl/network
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal, OneHotCategorical
import numpy as np

def layer_init(layer, w_scale=1.0):
  if hasattr(layer, 'weight') and len(layer.weight.shape) > 1:
    nn.init.orthogonal_(layer.weight.data)
    layer.weight.data.mul_(w_scale)
    nn.init.constant_(layer.bias.data, 0)
  return layer

class GELU(nn.Module):
  def forward(self, input):
    return F.gelu(input)


######################################################################################################
# Basic MLP, for non actor/critic things
######################################################################################################

class MLP(nn.Module):
  """Standard feedforward network.
  Args:
    input_size (int): number of input features
    output_size (int): number of output features
    hidden_sizes (int tuple): sizes of hidden layers

    norm: pre-activation module (e.g., nn.LayerNorm)
    activ: activation module (e.g., GELU, nn.ReLU)
    drop_prob: dropout probability to apply between layers (not applied to input)
  """
  def __init__(self, input_size, output_size=1, hidden_sizes=(256, 256), norm=nn.Identity, activ=nn.ReLU, drop_prob=0.):
    super().__init__()
    self.output_size = output_size

    layer_sizes = (input_size, ) + tuple(hidden_sizes) + (output_size, )
    if len(layer_sizes) == 2:
      layers = [nn.Linear(layer_sizes[0], layer_sizes[1], bias=False)]
    else:
      layers = []
      for dim_in, dim_out in zip(layer_sizes[:-1], layer_sizes[1:]):
        layers.append(nn.Linear(dim_in, dim_out))
        if norm not in [None, nn.Identity]:
          layers.append(norm(dim_out))
        layers.append(activ())
        if drop_prob > 0.:
          layers.append(nn.Dropout(p=drop_prob))
      layers = layers[:-(1 + (norm not in [None, nn.Identity]) + (drop_prob > 0))]
      layers = list(map(layer_init, layers))
    self.f = nn.Sequential(*layers)

  def forward(self, x):
    return self.f(x)
    
######################################################################################################
# Bodies. These are feature extractors; should be combined with a Head 
######################################################################################################


class NatureConvBody(nn.Module):
  def __init__(self, in_channels=4, feature_dim=512, norm = nn.Identity, activ = GELU):
    """Default in_channels is 4 because that is how many B/W Atari frames we stack"""
    super(NatureConvBody, self).__init__()
    self.feature_dim = feature_dim
    layers = [
        nn.Conv2d(in_channels, 32, kernel_size=8, stride=4),
        norm(32), activ(),
        nn.Conv2d(32, 64, kernel_size=4, stride=2),
        norm(64), activ(),
        nn.Conv2d(64, 64, kernel_size=3, stride=1),
        norm(64), activ(),
        nn.modules.Flatten(),
        nn.Linear(7 * 7 * 64, feature_dim),
        norm(feature_dim), activ(),
    ]
    layers = list(map(layer_init, layers))
    self.f = nn.Sequential(*layers)

  def forward(self, x):
    return self.f(x)


class FCBody(nn.Module):
  def __init__(self, input_size, layer_sizes=(256, 256), norm = nn.Identity, activ = GELU, use_layer_init = True):
    super(FCBody, self).__init__()
    self.feature_dim = layer_sizes[-1]

    layer_sizes = (input_size, ) + tuple(layer_sizes)
    layers = []
    for dim_in, dim_out in zip(layer_sizes[:-1], layer_sizes[1:]):
      layers += [
          nn.Linear(dim_in, dim_out),
          norm(dim_out), activ(),
      ]
    if use_layer_init:
      layers = list(map(layer_init, layers))
    self.f = nn.Sequential(*layers)

  def forward(self, x):
    return self.f(x)


######################################################################################################
# Main networks. Should be initialized with a body 
######################################################################################################


class Actor(nn.Module):
  """Returns batch of (action_dim,) vectors scaled to (-1, 1) (with tanh)"""
  def __init__(self, body : nn.Module, action_dim : int, max_action: float):
    super().__init__()
    self.body = body
    self.fc = layer_init(nn.Linear(self.body.feature_dim, action_dim))
    self.max_action = max_action

  def forward(self, x):
    return self.max_action * torch.tanh(self.fc(self.body(x)))

class StochasticActor(nn.Module):
  """Used by SAC"""
  def __init__(self, body : nn.Module, action_dim : int, max_action: float, log_std_bounds = (-20, 2)):
    super().__init__()
    self.body = body
    self.fc = layer_init(nn.Linear(self.body.feature_dim, action_dim*2))
    self.max_action = max_action
    self.min_log_std, self.max_log_std = log_std_bounds
    self.log2 = np.log(2)

  def forward(self, x):
    mu_and_log_std = self.fc(self.body(x))
    mu, log_std = torch.chunk(mu_and_log_std, 2, -1)

    action = mu
    logp_action = None
    if self.training:
      log_std = torch.clamp(log_std, self.min_log_std, self.max_log_std)
      std = torch.exp(log_std)
      action_distribution = Normal(mu, std)
      action = action_distribution.rsample()
      logp_action = action_distribution.log_prob(action).sum(axis=-1, keepdims=True)
      logp_action -= (2*(self.log2 - action - F.softplus(-2*action))).sum(axis=1, keepdims=True)

    action = torch.tanh(action)
    return self.max_action * action, logp_action

class Critic(nn.Module):
  """
  Can be used for DQN (output_dim = N) or DDPG (output_dim = 1),
  or a DPPG ensemble (output_dim = N), and so on.
  """
  def __init__(self, body : nn.Module, output_dim : int, use_layer_init : bool = True):
    super().__init__()
    self.body = body
    if use_layer_init:
      self.fc = layer_init(nn.Linear(self.body.feature_dim, output_dim))
    else:
      self.fc = nn.Linear(self.body.feature_dim, output_dim)

  def forward(self, *x):
    return self.fc(self.body(torch.cat(x, -1)))

class DuelingNet(nn.Module):
  def __init__(self, action_dim, body):
    super(DuelingNet, self).__init__()
    self.fc_value = nn.Linear(body.feature_dim, 1)
    self.fc_advantage = nn.Linear(body.feature_dim, action_dim)
    self.body = body

  def forward(self, x):
    phi = self.body(x)
    value = self.fc_value(phi)
    advantange = self.fc_advantage(phi)
    q = value.expand_as(advantange) + (advantange - advantange.mean(1, keepdim=True).expand_as(advantange))
    return q


class CategoricalNet(nn.Module):
  """For C51"""

  def __init__(self, action_dim, num_atoms, body):
    super(CategoricalNet, self).__init__()
    self.fc_categorical = nn.Linear(body.feature_dim, action_dim * num_atoms)
    self.action_dim = action_dim
    self.num_atoms = num_atoms
    self.body = body

  def forward(self, x):
    phi = self.body(x)
    pre_prob = self.fc_categorical(phi).view((-1, self.action_dim, self.num_atoms))
    prob = F.softmax(pre_prob, dim=-1)
    log_prob = F.log_softmax(pre_prob, dim=-1)
    return prob, log_prob


class QuantileNet(nn.Module):
  """For QR DQN"""

  def __init__(self, action_dim, num_quantiles, body):
    super(QuantileNet, self).__init__()
    self.fc_quantiles = nn.Linear(body.feature_dim, action_dim * num_quantiles)
    self.action_dim = action_dim
    self.num_quantiles = num_quantiles
    self.body = body

  def forward(self, x):
    phi = self.body(x)
    quantiles = self.fc_quantiles(phi)
    quantiles = quantiles.view((-1, self.action_dim, self.num_quantiles))
    return quantiles

######################################################################################################
# MDN for condition distribution estimation
######################################################################################################

# From https://github.com/tonyduan/mdn/blob/master/mdn/models.py 
class MixtureDensityNetwork(nn.Module):
  """
  Mixture density network.
  [ Bishop, 1994 ]
  Parameters
  ----------
  dim_in: int; dimensionality of the covariates
  dim_out: int; dimensionality of the response variable
  n_components: int; number of components in the mixture model
  """
  def __init__(self, pi_body: nn.Module, normal_body: nn.Module, output_dim: int, n_components: int):
    super().__init__()
    self.pi_network = CategoricalNetwork(pi_body, n_components)
    self.normal_network = MixtureDiagNormalNetwork(normal_body, output_dim, n_components)

  def forward(self, x):
    return self.pi_network(x), self.normal_network(x)

  def loss(self, x, y):
    pi, normal = self.forward(x)
    loglik = normal.log_prob(y.unsqueeze(1).expand_as(normal.loc))
    loglik = torch.sum(loglik, dim=2)
    loss = -torch.logsumexp(torch.log(pi.probs) + loglik, dim=1)
    return loss

class MixtureDiagNormalNetwork(nn.Module):
  def __init__(self, body: nn.Module, output_dim: int, n_components: int):
    super().__init__()
    self.body = body
    self.n_components = n_components
    self.fc = layer_init(nn.Linear(self.body.feature_dim, 2 * output_dim * n_components))

  def forward(self, x):
    params = self.fc(self.body(x))
    mean, sd = torch.split(params, params.shape[1] // 2, dim=1)
    mean = torch.stack(mean.split(mean.shape[1] // self.n_components, 1))
    sd = torch.stack(sd.split(sd.shape[1] // self.n_components, 1))
    return Normal(mean.transpose(0, 1), torch.exp(sd).transpose(0, 1))

class CategoricalNetwork(nn.Module):
  def __init__(self, body: nn.Module, n_components: int):
    super().__init__()
    self.body = body
    self.fc = layer_init(nn.Linear(self.body.feature_dim, n_components))

  def forward(self, x):
    params = self.fc(self.body(x))
    return OneHotCategorical(logits=params)



from .fusion import FusionClassifier
from .fusion import FusionRegressor
from .voting import VotingClassifier
from .voting import VotingRegressor
from .bagging import BaggingClassifier
from .bagging import BaggingRegressor
from .gradient_boosting import GradientBoostingClassifier
from .gradient_boosting import GradientBoostingRegressor
from .snapshot_ensemble import SnapshotEnsembleClassifier
from .snapshot_ensemble import SnapshotEnsembleRegressor
from .adversarial_training import AdversarialTrainingClassifier
from .adversarial_training import AdversarialTrainingRegressor
from .fast_geometric import FastGeometricClassifier
from .fast_geometric import FastGeometricRegressor
from .soft_gradient_boosting import SoftGradientBoostingClassifier
from .soft_gradient_boosting import SoftGradientBoostingRegressor


__all__ = [
    "FusionClassifier",
    "FusionRegressor",
    "VotingClassifier",
    "VotingRegressor",
    "BaggingClassifier",
    "BaggingRegressor",
    "GradientBoostingClassifier",
    "GradientBoostingRegressor",
    "SnapshotEnsembleClassifier",
    "SnapshotEnsembleRegressor",
    "AdversarialTrainingClassifier",
    "AdversarialTrainingRegressor",
    "FastGeometricClassifier",
    "FastGeometricRegressor",
    "SoftGradientBoostingClassifier",
    "SoftGradientBoostingRegressor",
]



from itertools import product

import numpy as np
import pandas as pd
from rich.console import Console

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.analytic import *
from rrt_ml.utilities.configs import *


cfg = MasterConfig().copy(deep=True)
cfg.env.general.seed = 0
cfg.env.general.gui = False
cfg.env.general.stats = True
cfg.env.general.max_timestep = 1000
cfg.env.target.pose = 1.5
cfg.env.car.pose = [0, 0, 0]
cfg.env.car.f_val = 240

# Initialize env and reset master
env = CarNavigationBulletEnv(cfg)
env.reset_master()

# Initialize dataframe
df = pd.DataFrame(columns=['nome', 'tempo', 'distancia', 'sucessos'])

# Params to search
l_weight_angle = [0.01, 0.05, 0.20, 0.40]
l_curvature = [2.5, 2.4, 2.3, 2.2]

# Logger
console = Console(width=150)

# it = product(l_weight_angle, l_step_size_mpc, l_n_horizon, l_curvature)]
i = 0
for i, params in enumerate(product(l_weight_angle, l_curvature)):

    # Change config
    cfg.rrt.mpc.weight_angle = params[0]
    cfg.rrt.rs.curvature = params[1]

    # Initialize controller
    rsmpc = MPCRS(cfg)

    # Set name
    nome = f'weight_angle:{params[0]}/curvature:{params[1]}'

    # Initialize sums
    distancia = 0
    tempo = 0
    n_success = 0

    # Log
    console.print(
            f"\n[cyan bold underline]Testing config:[/][cyan]{nome}[/]"
    )

    # Reset to keep episodes equal
    env.reset_master()

    # Get episodes infos
    for ep in range(10):

        # Control loop
        done, obs, info = False, env.reset(), None
        while not done:
            action = rsmpc.get_action(obs)
            obs, reward, done, info = env.step(action)

        # Log
        console.print(
            f"[cyan bold underline]Episode:[/] [cyan]{ep}[/]"
        )

        # Only add info on success
        if info['done_info']['success']:

            # Track num of success to take mean
            n_success += 1

            # Get length and time
            length = env.stats.get_distance_traveled(ep)
            time = env.stats.get_time_to_reach(ep)

            # Add to sum
            distancia += length
            tempo += time

    console.print(
        f"[red bold underline]Done config:[/][red]{i}[/]"
    )

    # Now we take the mean
    if n_success > 0:
        distancia /= n_success
        tempo /= n_success
    else:
        distancia = np.inf
        tempo = -np.inf

    # Add to dataframe
    df = pd.concat([pd.DataFrame([[nome, tempo, distancia, n_success]], columns=df.columns), df], ignore_index=True)

    # Save to file
    df.to_csv('mpc_tuning.csv')




# Adapted from: https://github.com/wjy5446/pytorch-Real-NVP

import torch
import torch.nn as nn
import numpy as np

class Loss(nn.Module):
    def __init__(self, prior):
        super(Loss, self).__init__()
        self.prior = prior

    def __call__(self, z, sum_log_det_jacobians):
        log_p = self.prior.log_prob(z)
        return -(log_p + sum_log_det_jacobians).mean()


class RealNVP(nn.Module):
    def __init__(self, prior=None, input_channel=2, lr=1e-3, num_layer_pairs=3, dev=None):
        super().__init__()

        if dev is None:
          self.dev = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
          self.dev = dev

        layers = []

        self.input_channel = input_channel

        for _ in range(num_layer_pairs):
          layers += [CouplingLayer("01", self.input_channel),
                     CouplingLayer("10", self.input_channel)]

        self.layers = nn.Sequential(*layers).to(self.dev)
        self.optimizer = torch.optim.Adam(self.parameters(), lr=lr)

        if prior is None:
            self.prior = torch.distributions.MultivariateNormal(torch.zeros(input_channel).to(self.dev),
                                                       torch.eye(input_channel).to(self.dev))
        else:
            self.prior = prior

    def forward(self, x, reverse=False):
        if not reverse:
            sum_log_det_jacobians = x.new_zeros(x.size(0))

            z = x
            for layer in self.layers:
              z, log_det_jacobians = layer(z, reverse)
              sum_log_det_jacobians += log_det_jacobians

            return z, sum_log_det_jacobians
        else:
            output = x
            for layer in list(self.layers)[::-1]:
              output = layer(output, reverse)

            return output

    def fit(self, data, epochs=10):
        if isinstance(data, torch.Tensor):
            data = data.to(self.dev)
        else:
            data = torch.from_numpy(data).to(self.dev)
        loss_log_det_jacobians = Loss(self.prior)
        for epoch in range(epochs):
            self.train()

            z, sum_log_det_jacobian = self(data)
            loss = loss_log_det_jacobians(z, sum_log_det_jacobian)

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

    def sample(self, num_samples=1000):
        z = self.prior.sample((num_samples,))
        x = self(z, reverse=True)
        return x.detach().cpu().numpy()

    def score_samples(self, x):
        z, sum_log_det_jacobian = self(torch.from_numpy(x).to(self.dev))

        log_pz = self.prior.log_prob(z)

        log_px = log_pz + sum_log_det_jacobian

        log_px = log_px.detach().cpu().numpy()

        return log_px


class CouplingLayer(nn.Module):
    def __init__(self, mask_type, input_channel, dev = None):
        super().__init__()
        self.function_s_t = Function_s_t(input_channel)
        self.mask_type = mask_type
        self.input_channel = input_channel

        if dev is None:
          self.dev = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
          self.dev = dev

        d = self.input_channel // 2
        if '01' in self.mask_type:
            self.mask = torch.tensor([[0.0]*d + (self.input_channel - d)*[1.0]]).to(self.dev)
        else:
            self.mask = torch.tensor([[1.0]*d + (self.input_channel - d)*[0.0]]).to(self.dev)

    def forward(self, x, reverse=False):

        if not reverse:
            # masked half of x
            x1 = x * self.mask
            s,t = self.function_s_t(x1, self.mask)

            # z_1:d = x_1:d
            # z_d+1:D = exp(s(x_1:d)) * x_d+1:D + m(x_1:d)
            y = x1 + ((-self.mask+1.0) * (x*torch.exp(s)+t))

            # calculation of jacobians
            log_det_jacobian = torch.sum(s, 1)

            return y, log_det_jacobian
        else:
            # masked half of y
            x1 = x * self.mask
            s,t = self.function_s_t(x1, self.mask)

            # x_1:d = z_1:d
            # x_d+1:D = z_d+1:D - m(z_1:d) * exp(-s(z_1:d))
            y = x1 + (-self.mask+1.0) * ((x-t) * torch.exp(-s))

            return y

class Function_s_t(nn.Module):
    ############################################
    # scale, translation function
    ############################################
    def __init__(self, input_channel, channel=256):
        super().__init__()
        self.input_channel = input_channel
        layers = []

        layers += [
            nn.Linear(input_channel, channel),
            nn.LeakyReLU(),
            nn.Linear(channel, channel),
            nn.LeakyReLU(),
            nn.Linear(channel, input_channel*2)]

        self.model = nn.Sequential(*layers)
        self.w_scale = torch.rand(1, requires_grad=True)

    def forward(self, x, mask):
        x = self.model(x)

        #######################################
        # scale function : first half dimension
        # translation function : second half dimension
        #######################################
        s = x[:,:self.input_channel] * (-mask+1)
        t = x[:,self.input_channel:] * (-mask+1)

        s = nn.Tanh()(s)

        return s, t